/* ############################# MPC License ############################## */
/* # Wed Nov 19 15:19:19 CET 2008                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # IDDN.FR.001.230040.000.S.P.2007.000.10000                            # */
/* # This file is part of the MPC Runtime.                                # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - PERACHE Marc marc.perache@cea.fr                                 # */
/* #                                                                      # */
/* ######################################################################## */
#include "sctk_low_level_comm.h"
#include "sctk_debug.h"
#include "sctk_thread.h"
#include "sctk_kernel_thread.h"
#include "sctk_rpc.h"
#include "sctk_elan4.h"
#include <math.h>
#include <sched.h>
#include <sctk_asm.h>

#ifdef MPC_USE_ELAN
#include "sctk_iso_alloc.h"
#include "sctk_math.h"


static int sctk_rail_number = 0;

ELAN_BASE *elan_base;
static char sctk_elan_network_mode[4096];
static size_t sctk_elan_alloc_mem_used = 0;

typedef ELAN_EVENT *sctk_rdma_specific_t;
static void sctk_net_ptp_poll_elan (void *arg);

/******************************************************************/
/* LOG METHOD                                                     */
/******************************************************************/
#ifdef ELAN_LOG
#warning "ELAN LOG ENABLE"
static FILE *elan_log_file = NULL;
#endif

static void
init_elan_display ()
{
  sprintf (sctk_elan_network_mode, "Elan_%s (%dKo)", elan_version (),
	   (int) ((double) sctk_elan_alloc_mem_used / 1024.0));
  sctk_network_mode = sctk_elan_network_mode;
#if 0
  sctk_debug ("ELAN VERSION %s", elan_version ());
  sctk_debug ("%d rails", elan_nRails (elan_base->state));
  sctk_debug ("elan_base->putget_smallputsize %d",
	      (int) elan_base->putget_smallputsize);
  sctk_debug ("elan_base->putget_stripeputsize %d",
	      (int) elan_base->putget_stripeputsize);
  sctk_debug ("elan_base->putget_stripegetsize %d",
	      (int) elan_base->putget_stripegetsize);
  sctk_debug ("elan_base->putget_throttle %d",
	      (int) elan_base->putget_throttle);
  sctk_debug ("Elan memory %dMo",
	      (int) elan_base->gallocElan_size / (1024 * 1024));
  sctk_debug ("Main memory %dMo",
	      (int) elan_base->galloc_size / (1024 * 1024));
#endif

}

/******************************************************************/
/* RPC                                                            */
/******************************************************************/
static void **sctk_rpc_buffer;
static sctk_rpc_t *volatile sctk_send_rpc_used;
static sctk_rpc_t *volatile sctk_recv_rpc_used;
static ELAN_PGCTRL *PGCTRL_RPC;
static void
sctk_net_adm_poll_elan (void *arg)
{
  /*RPC deamon */
  int i;
  for (i = 0; i < sctk_process_number; i++)
    {
      if (sctk_recv_rpc_used[i])
	{
	  void *tmp_arg;
	  sctk_rpc_t func;

	  sctk_nodebug ("RPC recv message form %d func %p %p", i,
			sctk_recv_rpc_used[i], sctk_rpc_buffer[i]);
	  tmp_arg = sctk_rpc_get_slot (sctk_rpc_buffer[i]);
	  func = sctk_recv_rpc_used[i];

	  sctk_recv_rpc_used[i] = NULL;
	  elan_wait (elan_doput (PGCTRL_RPC, &(sctk_recv_rpc_used[i]),
				 &(sctk_send_rpc_used[sctk_process_rank]),
				 (ELAN_ADDR) NULL, sizeof (sctk_rpc_t), i,
				 ELAN_RAIL_ALL), ELAN_WAIT_EVENT);


	  sctk_rpc_execute (func, tmp_arg);
	  sctk_nodebug ("RPC recv message form %d done", i);
	}
    }
}

static void
sctk_net_rpc_driver (void (*func) (void *), int destination, void *arg,
		     size_t arg_size)
{
  static sctk_thread_mutex_t lock = SCTK_THREAD_MUTEX_INITIALIZER;
  ELAN_EVENT *event;
  sctk_thread_mutex_lock (&lock);
  sctk_nodebug ("RPC to %d %p wait free slot", destination, func);
  sctk_thread_wait_for_value ((int *) &(sctk_send_rpc_used[destination]), 0);
  while (sctk_send_rpc_used[destination])
    {
      sctk_net_ptp_poll_elan (NULL);
      sctk_thread_yield ();
    }

  sctk_nodebug ("RPC to %d %p(%p) send (%p)", destination, func, arg,
		sctk_rpc_buffer[destination]);
  elan_wait (elan_doput
	     (PGCTRL_RPC, arg, sctk_rpc_buffer[sctk_process_rank],
	      (ELAN_ADDR) NULL, arg_size, destination, ELAN_RAIL_ALL),
	     ELAN_WAIT_EVENT);

  sctk_send_rpc_used[destination] = func;
  sctk_nodebug ("RPC to %d %p confirmation (%p)", destination, func,
		&(sctk_recv_rpc_used[destination]));
  event =
    elan_doput (PGCTRL_RPC, &(sctk_send_rpc_used[destination]),
		&(sctk_recv_rpc_used[sctk_process_rank]), (ELAN_ADDR) NULL,
		sizeof (sctk_rpc_t), destination, ELAN_RAIL_ALL);
  elan_wait (event, ELAN_WAIT_EVENT);

  sctk_nodebug ("RPC to %d %p all done", destination, func);
  sctk_thread_mutex_unlock (&lock);
}

static void
sctk_net_rpc_retrive_driver (void *dest, void *src, size_t arg_size,
			     int process, int *ack)
{
  static sctk_thread_mutex_t lock = SCTK_THREAD_MUTEX_INITIALIZER;
  ELAN_EVENT *event;
  static int done = 1;
  sctk_thread_mutex_lock (&lock);
  event = elan_doget (PGCTRL_RPC,
		      src, dest, arg_size, process, ELAN_RAIL_ALL);
  elan_wait (event, ELAN_WAIT_EVENT);

  elan_wait (elan_doput (PGCTRL_RPC, &(done),
			 ack,
			 (ELAN_ADDR) NULL, sizeof (int), process,
			 ELAN_RAIL_ALL), ELAN_WAIT_EVENT);
  sctk_thread_mutex_unlock (&lock);
}

static void
sctk_net_rpc_send_driver (void *dest, void *src, size_t arg_size, int process,
			  int *ack)
{
  static sctk_thread_mutex_t lock = SCTK_THREAD_MUTEX_INITIALIZER;
  ELAN_EVENT *event;
  static int done = 1;
  sctk_thread_mutex_lock (&lock);
  if (arg_size)
    {
      sctk_nodebug ("Send message");
      event = elan_doput (PGCTRL_RPC,
			  src, dest,
			  (ELAN_ADDR) NULL, arg_size, process, ELAN_RAIL_ALL);
      elan_wait (event, ELAN_WAIT_EVENT);
    }
  sctk_nodebug ("ACK");
  elan_wait (elan_doput (PGCTRL_RPC, &(done),
			 ack,
			 (ELAN_ADDR) NULL, sizeof (int), process,
			 ELAN_RAIL_ALL), ELAN_WAIT_EVENT);
  sctk_thread_mutex_unlock (&lock);
}

/******************************************************************/
/* COLLECTIVES                                                    */
/******************************************************************/
#define collective_buffer_size (1024)
static ELAN_PGCTRL *PGCTRL_COLLECTIVE;
#define collective_op_steps 32

typedef struct
{
  volatile int reinit;
  volatile int done;
  volatile char *addr;
  volatile int src;
  volatile int nputs;
} sctk_collective_intern_msg_t;

typedef struct
{
  volatile ELAN_EVENT *event_rec[SCTK_MAX_RAIL];
  void *devent;
} sctk_collective_intern_devent_t;

typedef struct
{
  sctk_collective_intern_msg_t msg[collective_op_steps];
  sctk_collective_intern_devent_t devent[collective_op_steps];
  ELAN_EVENT *event;
  char buffer[collective_op_steps][collective_buffer_size];
} sctk_collective_intern_t;

typedef struct
{
  sctk_collective_intern_t collective[2];
  int val;
  int step_number;
} sctk_collective_t;
static sctk_collective_t sctk_bar[SCTK_MAX_COMMUNICATOR_NUMBER];

static inline void
sctk_elan_collective_reinit_all_parts (sctk_collective_intern_msg_t * coll)
{
  coll->done = -1;
  coll->reinit = -1;
  coll->addr = (void *) -1;
  coll->src = -1;
  coll->nputs = -1;
}

static inline void
sctk_elan_collective_init_all_parts (sctk_collective_intern_msg_t * coll,
				     int done, int reinit, void *addr,
				     int nputs)
{
  coll->done = done;
  coll->reinit = reinit;
  coll->addr = addr;
  coll->src = sctk_process_rank;
  coll->nputs = nputs;
}

static void
sctk_collective_init (sctk_collective_t * bar, int nb)
{
  int i;
  bar->val = 0;
  bar->step_number = (int) ceil (log2 (nb));
  assume (bar->step_number <= collective_op_steps);
  for (i = 0; i < bar->step_number; i++)
    {
      int j;
      sctk_elan_collective_reinit_all_parts (&(bar->collective[0].msg[i]));
      sctk_elan_collective_reinit_all_parts (&(bar->collective[1].msg[i]));

      bar->collective[0].devent[i].devent =
	elan_allocElan (elan_base->state, ELAN_EVENT_ALIGN,
			sizeof (ELAN_EVENT_ELAN));
      assume (bar->collective[0].devent[i].devent != NULL);
      bar->collective[1].devent[i].devent =
	elan_allocElan (elan_base->state, ELAN_EVENT_ALIGN,
			sizeof (ELAN_EVENT_ELAN));
      assume (bar->collective[1].devent[i].devent != NULL);
      sctk_nodebug ("DEVENTS %p %p in %p %p",
		    bar->collective[0].devent[i].devent,
		    bar->collective[1].devent[i].devent,
		    &(bar->collective[0].devent[i]),
		    &(bar->collective[1].devent[i]));

      for (j = 0; j < sctk_rail_number; j++)
	{
	  bar->collective[0].devent[i].event_rec[j] =
	    elan_initEvent (elan_base->state, elan_base->state->rail[j],
			    (ELAN_ADDR) bar->collective[0].devent[i].devent,
			    1);
	  bar->collective[1].devent[i].event_rec[j] =
	    elan_initEvent (elan_base->state, elan_base->state->rail[j],
			    (ELAN_ADDR) bar->collective[1].devent[i].devent,
			    1);
	  sctk_nodebug ("%p %p ", bar->collective[0].devent[i].event_rec[j],
			bar->collective[1].devent[i].event_rec[j]);
	}
    }
}

static void
sctk_elan_poll_collective (sctk_collective_intern_t * bar)
{
  if (bar->event)
    {
      if (elan_poll (bar->event, 0))
	{
	  bar->event = NULL;
	}
    }
}

static inline void
sctk_elan_collective_wait_all_parts (sctk_collective_intern_msg_t * coll,
				     sctk_collective_intern_devent_t * devent)
{
  int j;
  while ((coll->reinit == -1) || (coll->addr == (void *) -1)
	 || (coll->src == -1) || (coll->nputs == -1))
    {
      sctk_net_ptp_poll_elan (NULL);
      sctk_thread_yield ();
    }

  sctk_nodebug ("nputs %d", coll->nputs);
  if (coll->nputs > 1)
    {
      for (j = 0; j < sctk_rail_number; j++)
	{
	  elan_issueWaitevent ((ELAN_EVENT *) devent->event_rec[j],
			       coll->nputs - 1);
/*       _elan_deventDump("REINIT ",(ELAN_EVENT*)devent->event_rec[j]); */
	}
    }

  sctk_nodebug ("devent %p %p", devent->devent, devent);
  for (j = 0; j < sctk_rail_number; j++)
    {
      sctk_nodebug ("Event %p %d", devent->event_rec[j], j);
/*     _elan_deventDump("WAIT   ",(ELAN_EVENT*)devent->event_rec[j]); */
      elan_wait ((ELAN_EVENT *) devent->event_rec[j], ELAN_WAIT_EVENT);
    }
}


/* BARRIER */
static inline void
sctk_elan_barrier_all_processes (sctk_collective_communications_t * com,
				 sctk_collective_t * bar_s)
{
  int i;
  int m;
  int acc = 1;
  int next;
  sctk_collective_intern_msg_t msg;
  sctk_collective_intern_t *bar;
  int process_rank;
  int process_number;
  int mode = 1;
  int reinit;

  if (com->involved_nb != sctk_process_number)
    {
      mode = 0;
    }

  if (mode)
    {
      m = bar_s->step_number;
      process_number = sctk_process_number;
      process_rank = sctk_process_rank;
    }
  else
    {
      int j;
      m = (int) ceil (log2 (com->involved_nb));
      process_number = com->involved_nb;
      process_rank = 0;
      for (j = 0; j < sctk_process_rank; j++)
	{
	  if (com->involved_list[j])
	    {
	      process_rank++;
	    }
	}
    }

  bar = &(bar_s->collective[bar_s->val]);
  bar_s->val = (bar_s->val + 1) % 2;

  sctk_elan_collective_init_all_parts (&msg, 0, com->reinitialize, NULL, 1);

  if (mode)
    {
      next = (process_rank + acc) % process_number;
    }
  else
    {
      int tmp_next;
      int j;
      tmp_next = (process_rank + acc) % process_number;
      next = 0;
      while (!com->involved_list[next])
	next++;
      for (j = 0; j < tmp_next; j++)
	{
	  next++;
	  while (!com->involved_list[next])
	    next++;
	}
    }

  for (i = 0; i < m; i++)
    {
      msg.done = i;
      sctk_nodebug ("devent %p", bar->devent[i].devent);
      bar->event =
	elan_doput (PGCTRL_COLLECTIVE, &msg, &(bar->msg[i]),
		    (ELAN_ADDR) bar->devent[i].devent,
		    sizeof (sctk_collective_intern_msg_t), next,
		    ELAN_RAIL_ALL);

      acc *= 2;
      if (mode)
	{
	  next = (process_rank + acc) % process_number;
	}
      else
	{
	  int tmp_next;
	  int j;
	  tmp_next = (process_rank + acc) % process_number;
	  next = 0;
	  while (!com->involved_list[next])
	    next++;
	  for (j = 0; j < tmp_next; j++)
	    {
	      next++;
	      while (!com->involved_list[next])
		next++;
	    }
	}


      if (bar->event)
	{
	  elan_wait (bar->event, ELAN_WAIT_EVENT);
	  bar->event = NULL;
	}

      sctk_thread_wait_for_value_and_poll ((int *) &(bar->msg[i].done), i,
					   (void (*)(void *))
					   sctk_elan_poll_collective, bar);



      sctk_elan_collective_wait_all_parts (&(bar->msg[i]), &(bar->devent[i]));

      reinit = msg.reinit;
      msg.reinit = reinit | bar->msg[i].reinit;

      sctk_elan_collective_reinit_all_parts (&(bar->msg[i]));
    }
  com->reinitialize = msg.reinit;
}

static inline void
sctk_elan_barrier (sctk_collective_communications_t * com,
		   sctk_virtual_processor_t * my_vp)
{
#ifdef ELAN_LOG
  sctk_log (elan_log_file, "barrier");
#endif

  SCTK_PROFIL_START (sctk_elan_barrier);
  sctk_nodebug ("Perform barrier on %d", com->id);
  sctk_elan_barrier_all_processes (com, &(sctk_bar[com->id]));
/*   elan_gsync(elan_base->allGroup);  */
  SCTK_PROFIL_END (sctk_elan_barrier);
}

/* BROADCAST */

static inline void
sctk_elan_broadcast_all_processes (sctk_collective_communications_t * com,
				   sctk_virtual_processor_t * my_vp,
				   sctk_collective_t * bar_s,
				   const size_t elem_size,
				   const size_t nb_elem)
{
  int i;
  int m;
  int acc = 1;
  int next;
  sctk_collective_intern_msg_t msg;
  sctk_collective_intern_t *bar;
  int have_data = 0;
  int process_rank;
  int process_number;
  int reinit;
  int mode = 1;

  if (com->involved_nb != sctk_process_number)
    {
      mode = 0;
    }

  if (mode)
    {
      m = bar_s->step_number;
      process_number = sctk_process_number;
      process_rank = sctk_process_rank;
    }
  else
    {
      int j;
      m = (int) ceil (log2 (com->involved_nb));
      process_number = com->involved_nb;
      process_rank = 0;
      for (j = 0; j < sctk_process_rank; j++)
	{
	  if (com->involved_list[j])
	    {
	      process_rank++;
	    }
	}
    }

  bar = &(bar_s->collective[bar_s->val]);
  bar_s->val = (bar_s->val + 1) % 2;

  sctk_nodebug ("INIT IN (%p) |%s| OUT %p %d",
		my_vp->data.data_in, my_vp->data.data_in,
		my_vp->data.data_out, elem_size * nb_elem);

  sctk_elan_collective_init_all_parts (&msg, 0, com->reinitialize,
				       my_vp->data.data_in, 1);
  if (my_vp->data.data_in)
    {
      have_data = 1;
    }

  if (mode)
    {
      next = (process_rank + acc) % process_number;
    }
  else
    {
      int tmp_next;
      int j;
      tmp_next = (process_rank + acc) % process_number;
      next = 0;
      while (!com->involved_list[next])
	next++;
      for (j = 0; j < tmp_next; j++)
	{
	  next++;
	  while (!com->involved_list[next])
	    next++;
	}
    }
  for (i = 0; i < m; i++)
    {
      msg.done = i;
      sctk_nodebug ("SEND MSG to %d : %p", next, msg.addr);

      if (!have_data)
	{
	  bar->event =
	    elan_doput (PGCTRL_COLLECTIVE, &msg, &(bar->msg[i]),
			(ELAN_ADDR) bar->devent[i].devent,
			sizeof (sctk_collective_intern_msg_t), next,
			ELAN_RAIL_ALL);
	}
      else
	{
	  if (elem_size * nb_elem <= collective_buffer_size)
	    {
	      msg.nputs++;
	      bar->event =
		elan_doput (PGCTRL_COLLECTIVE, (void *) msg.addr,
			    bar->buffer[0], (ELAN_ADDR) bar->devent[i].devent,
			    elem_size * nb_elem, next, ELAN_RAIL_ALL);

/* 	elan_wait(bar->event,ELAN_WAIT_EVENT); */

	      bar->event = elan_link (bar->event,
				      elan_doput (PGCTRL_COLLECTIVE, &msg,
						  &(bar->msg[i]),
						  (ELAN_ADDR) bar->devent[i].
						  devent,
						  sizeof
						  (sctk_collective_intern_msg_t),
						  next, ELAN_RAIL_ALL));
	    }
	  else
	    {
	      bar->event =
		elan_doput (PGCTRL_COLLECTIVE, &msg, &(bar->msg[i]),
			    (ELAN_ADDR) bar->devent[i].devent,
			    sizeof (sctk_collective_intern_msg_t), next,
			    ELAN_RAIL_ALL);
	    }
	}

      acc *= 2;
      if (mode)
	{
	  next = (process_rank + acc) % process_number;
	}
      else
	{
	  int tmp_next;
	  int j;
	  tmp_next = (process_rank + acc) % process_number;
	  next = 0;
	  while (!com->involved_list[next])
	    next++;
	  for (j = 0; j < tmp_next; j++)
	    {
	      next++;
	      while (!com->involved_list[next])
		next++;
	    }
	}

      if (bar->event)
	{
	  elan_wait (bar->event, ELAN_WAIT_EVENT);
	  bar->event = NULL;
	}

      sctk_thread_wait_for_value_and_poll ((int *) &(bar->msg[i].done), i,
					   (void (*)(void *))
					   sctk_elan_poll_collective, bar);

      sctk_elan_collective_wait_all_parts (&(bar->msg[i]), &(bar->devent[i]));

      reinit = msg.reinit;
      msg.reinit = reinit | bar->msg[i].reinit;

      /*Broadcast core */
      if (!have_data && bar->msg[i].addr)
	{
	  if (elem_size * nb_elem <= collective_buffer_size)
	    {
	      memcpy (my_vp->data.data_out, bar->buffer[0],
		      elem_size * nb_elem);
	    }
	  else
	    {
	      sctk_nodebug ("RECV %d %p -> %p (%s)",
			    bar->msg[i].src, bar->msg[i].addr,
			    my_vp->data.data_out, my_vp->data.data_out);

	      elan_wait (elan_doget (PGCTRL_COLLECTIVE,
				     (void *) bar->msg[i].addr,
				     my_vp->data.data_out,
				     elem_size * nb_elem, bar->msg[i].src,
				     ELAN_RAIL_ALL), ELAN_WAIT_EVENT);

	      sctk_nodebug ("RECV %d %p -> %p (%s)",
			    bar->msg[i].src, bar->msg[i].addr,
			    my_vp->data.data_out, my_vp->data.data_out);
	      sctk_nodebug ("RECVED |%s|", my_vp->data.data_out);
	    }
	  msg.addr = my_vp->data.data_out;
	  have_data = 1;
	}

      sctk_elan_collective_reinit_all_parts (&(bar->msg[i]));
    }
  com->reinitialize = msg.reinit;

  if (my_vp->data.data_out && my_vp->data.data_in)
    {
      my_vp->data.data_out = my_vp->data.data_in;
    }

  if (elem_size * nb_elem > collective_buffer_size)
    {
      sctk_elan_barrier (com, my_vp);
    }
  sctk_nodebug ("sctk_elan_broadcast_all_processes done |%s||%s|",
		my_vp->data.data_in, my_vp->data.data_out);
}

static inline void
sctk_elan_broadcast (sctk_collective_communications_t * com,
		     sctk_virtual_processor_t * my_vp,
		     const size_t elem_size, const size_t nb_elem)
{
#ifdef ELAN_LOG
  sctk_log (elan_log_file, "broadcast");
#endif
  sctk_nodebug ("Perform broadcast on %d", com->id);
  sctk_elan_broadcast_all_processes (com, my_vp, &(sctk_bar[com->id]),
				     elem_size, nb_elem);
}

/* ALLREDUCE */
static inline void
sctk_allreduce (sctk_collective_communications_t * com,
		sctk_virtual_processor_t * my_vp,
		const size_t elem_size,
		const size_t nb_elem,
		void (*func) (const void *, void *, size_t,
			      sctk_datatype_t),
		const sctk_datatype_t data_type)
{
  int step_number;
  int i;
  int to;
  int from;
  int acc = 1;
  sctk_collective_intern_t *bar;
  sctk_collective_t *bar_s;
  int mode = 1;
  int process_rank;
  int process_number;
  sctk_collective_intern_msg_t msg;
  void *tmp_buf = NULL;
  int reinit;
  sctk_nodebug ("Perform reduce on %d", com->id);
#ifdef ELAN_LOG
  sctk_log (elan_log_file, "allreduce");
#endif

  if (com->involved_nb != sctk_process_number)
    {
      mode = 0;
    }

  bar_s = &(sctk_bar[com->id]);

  if (mode)
    {
      step_number = bar_s->step_number;
      process_number = sctk_process_number;
      process_rank = sctk_process_rank;
    }
  else
    {
      int j;
      step_number = (int) ceil (log2 (com->involved_nb));
      process_number = com->involved_nb;
      process_rank = 0;
      for (j = 0; j < sctk_process_rank; j++)
	{
	  if (com->involved_list[j])
	    {
	      process_rank++;
	    }
	}
    }

  if (elem_size * nb_elem > collective_buffer_size)
    {
      tmp_buf = sctk_malloc (elem_size * nb_elem);
    }


  bar = &(bar_s->collective[bar_s->val]);
  bar_s->val = (bar_s->val + 1) % 2;

  step_number = (int) ceil (log2 (com->involved_nb));

  sctk_nodebug ("in %d inout %d", ((int *) my_vp->data.data_in)[0],
		((int *) my_vp->data.data_out)[0]);
/*   memcpy(my_vp->data.data_out,my_vp->data.data_in,elem_size*nb_elem); */
/*   sctk_nodebug("in %d inout %d",((int*)my_vp->data.data_in)[0],((int*)my_vp->data.data_out)[0]); */

  sctk_elan_collective_init_all_parts (&msg, 0, com->reinitialize,
				       my_vp->data.data_out, 1);

  msg.done = process_rank;
  /*Binary tree */
  for (i = 0; i < step_number; i++)
    {
      msg.nputs = 1;
      to = process_rank & (~acc);
      if (!mode)
	{
	  int tmp_next;
	  int j;
	  tmp_next = to;
	  to = 0;
	  while (!com->involved_list[to])
	    to++;
	  for (j = 0; j < tmp_next; j++)
	    {
	      to++;
	      while (!com->involved_list[to])
		to++;
	    }
	}

      if (process_rank == to)
	{
	  from = process_rank + acc;
	  if (!mode)
	    {
	      int tmp_next;
	      int j;
	      tmp_next = from;
	      from = 0;
	      while (!com->involved_list[from])
		from++;
	      for (j = 0; j < tmp_next; j++)
		{
		  from++;
		  while (!com->involved_list[from])
		    from++;
		}
	    }
	  if (from < process_number)
	    {
	      sctk_nodebug ("wait msg from %d", from);
	      sctk_thread_wait_for_value_and_poll ((int *)
						   &(bar->msg[i].done), from,
						   (void (*)(void *))
						   sctk_elan_poll_collective,
						   bar);
	      sctk_elan_collective_wait_all_parts (&(bar->msg[i]),
						   &(bar->devent[i]));
	      reinit = msg.reinit;
	      msg.reinit = reinit | bar->msg[i].reinit;

	      /*Reduce operation */
	      if (elem_size * nb_elem <= collective_buffer_size)
		{
		  sctk_nodebug ("in %d inout %d", ((int *) bar->buffer[i])[0],
				((int *) msg.addr)[0]);
		  func (bar->buffer[i], (void *) msg.addr, nb_elem,
			data_type);
		  sctk_nodebug ("in %d inout %d", ((int *) bar->buffer[i])[0],
				((int *) msg.addr)[0]);
		}
	      else
		{
		  elan_wait (elan_doget (PGCTRL_COLLECTIVE,
					 (void *) bar->msg[i].addr,
					 tmp_buf,
					 elem_size * nb_elem, bar->msg[i].src,
					 ELAN_RAIL_ALL), ELAN_WAIT_EVENT);
		  func (tmp_buf, (void *) msg.addr, nb_elem, data_type);
		}

	      sctk_elan_collective_reinit_all_parts (&(bar->msg[i]));
	      sctk_nodebug ("wait msg from %d done", from);
	    }
	}
      else
	{
	  sctk_nodebug ("send msg to %d", to);
	  if (elem_size * nb_elem <= collective_buffer_size)
	    {
	      msg.nputs++;
	      bar->event =
		elan_doput (PGCTRL_COLLECTIVE, (void *) msg.addr,
			    bar->buffer[i], (ELAN_ADDR) bar->devent[i].devent,
			    elem_size * nb_elem, to, ELAN_RAIL_ALL);

	      bar->event = elan_link (bar->event,
				      elan_doput (PGCTRL_COLLECTIVE, &msg,
						  &(bar->msg[i]),
						  (ELAN_ADDR) bar->devent[i].
						  devent,
						  sizeof
						  (sctk_collective_intern_msg_t),
						  to, ELAN_RAIL_ALL));
	    }
	  else
	    {
	      bar->event =
		elan_doput (PGCTRL_COLLECTIVE, &msg, &(bar->msg[i]),
			    (ELAN_ADDR) bar->devent[i].devent,
			    sizeof (sctk_collective_intern_msg_t), to,
			    ELAN_RAIL_ALL);
	    }

	  elan_wait (bar->event, ELAN_WAIT_EVENT);
	  bar->event = NULL;

	  sctk_nodebug ("send msg to %d done", to);

	  sctk_nodebug ("wait msg from %d %d", to, i);
	  sctk_thread_wait_for_value_and_poll ((int *) &(bar->msg[i].done),
					       to,
					       (void (*)(void *))
					       sctk_elan_poll_collective,
					       bar);
	  sctk_elan_collective_wait_all_parts (&(bar->msg[i]),
					       &(bar->devent[i]));
	  sctk_nodebug ("wait msg from %d %d done", to, i);
	  reinit = msg.reinit;
	  msg.reinit = reinit | bar->msg[i].reinit;

	  /*Store result */
	  if (elem_size * nb_elem <= collective_buffer_size)
	    {
	      memcpy (my_vp->data.data_out, bar->buffer[i],
		      elem_size * nb_elem);
	    }
	  else
	    {
	      elan_wait (elan_doget (PGCTRL_COLLECTIVE,
				     (void *) bar->msg[i].addr,
				     tmp_buf,
				     elem_size * nb_elem, bar->msg[i].src,
				     ELAN_RAIL_ALL), ELAN_WAIT_EVENT);
	      memcpy (my_vp->data.data_out, tmp_buf, elem_size * nb_elem);
	    }

	  sctk_elan_collective_reinit_all_parts (&(bar->msg[i]));
	  break;
	}
      acc *= 2;
    }

  sctk_nodebug ("Broadcast phase");
  for (; i > 0; i--)
    {
      msg.nputs = 1;
      acc /= 2;
      to = process_rank + acc;
      if (!mode)
	{
	  int tmp_next;
	  int j;
	  tmp_next = to;
	  to = 0;
	  while (!com->involved_list[to])
	    to++;
	  for (j = 0; j < tmp_next; j++)
	    {
	      to++;
	      while (!com->involved_list[to])
		to++;
	    }
	}
      if (to < process_number)
	{
	  sctk_nodebug ("send msg to %d %d broadcast", to, i);
	  if (elem_size * nb_elem <= collective_buffer_size)
	    {
	      msg.nputs++;
	      bar->event =
		elan_doput (PGCTRL_COLLECTIVE, (void *) msg.addr,
			    bar->buffer[i - 1],
			    (ELAN_ADDR) bar->devent[i - 1].devent,
			    elem_size * nb_elem, to, ELAN_RAIL_ALL);

	      bar->event = elan_link (bar->event,
				      elan_doput (PGCTRL_COLLECTIVE, &msg,
						  &(bar->msg[i - 1]),
						  (ELAN_ADDR) bar->devent[i -
									  1].
						  devent,
						  sizeof
						  (sctk_collective_intern_msg_t),
						  to, ELAN_RAIL_ALL));
	    }
	  else
	    {
	      bar->event =
		elan_doput (PGCTRL_COLLECTIVE, &msg, &(bar->msg[i - 1]),
			    (ELAN_ADDR) bar->devent[i - 1].devent,
			    sizeof (sctk_collective_intern_msg_t), to,
			    ELAN_RAIL_ALL);
	    }

	  elan_wait (bar->event, ELAN_WAIT_EVENT);
	  bar->event = NULL;

	}
    }

  sctk_nodebug ("Reduce Broadcast phase done");
  com->reinitialize = msg.reinit;

  if (elem_size * nb_elem > collective_buffer_size)
    {
      sctk_free (tmp_buf);
      sctk_elan_barrier (com, my_vp);
      sctk_nodebug ("Final barrier done");
    }
}

static void
sctk_net_collective_op_driver (sctk_collective_communications_t * com,
			       sctk_virtual_processor_t * my_vp,
			       const size_t elem_size,
			       const size_t nb_elem,
			       void (*func) (const void *, void *, size_t,
					     sctk_datatype_t),
			       const sctk_datatype_t data_type)
{
  sctk_nodebug ("begin collective");
  if (nb_elem == 0)
    {
      sctk_elan_barrier (com, my_vp);
      sctk_nodebug ("end collective barrier");
      return;
    }
  else
    {
      if (func == NULL)
	{
	  sctk_elan_broadcast (com, my_vp, elem_size, nb_elem);
	  sctk_nodebug ("end collective broadcast");
	  return;
	}
      else
	{
	  sctk_allreduce (com, my_vp, elem_size, nb_elem, func, data_type);
	  sctk_nodebug ("end collective reduce");
	  return;
	}
    }
}

/******************************************************************/
/* PT 2 PT                                                        */
/******************************************************************/
#define ptp_max_put_nb 5
#define ptp_buffer_copy_size (2*1024)
#define ptp_buffer_put_size (8*1024)
#define ptp_buffer_size (1024*1024)
#define ptp_buffer_slots 16

/* static ELAN_EVENT *elan_doput_small(ELAN_PGCTRL *pgctrl, void *source, void *target,  */
/* 				    ELAN_ADDR devent, size_t len, u_int destvp, int rail){ */
/*   static sctk_spinlock_t spinlocks[SCTK_MAX_RAIL]; */
/*   ELAN_EVENT * tmp; */
/*   if(rail == ELAN_RAIL_ALL){ */
/*     tmp = elan_doput(pgctrl, source, target, devent, len, destvp, rail); */
/*   } else { */
/*     sctk_spinlock_lock(&(spinlocks[rail])); */
/*     tmp = elan_doput(pgctrl, source, target, devent, len, destvp, rail); */
/*     sctk_spinlock_unlock(&(spinlocks[rail])); */
/*   } */
/*   return tmp; */
/* } */

typedef enum
{
  ptp_copy_type = 0, ptp_put_type = 1, ptp_get_type = 2, ptp_copy_put_type =
    3, ptp_nothing_type = -1
} elan_ptp_message_type_t;

static inline char *sctk_elan_print_message_type (elan_ptp_message_type_t
						  result);

static inline elan_ptp_message_type_t
sctk_elan_determine_message_type (size_t size,
				  sctk_thread_ptp_message_t * msg)
{
  elan_ptp_message_type_t result;
  if (size <= ptp_buffer_size)
    {
      if (size <= ptp_buffer_copy_size)
	{
	  /* Very small messages */
	  result = ptp_copy_type;
	}
      else
	{
	  if (msg->message.max_count * msg->message.nb_items <= 1)
	    {
	      /* Simple messages */
	      result = ptp_put_type;

	      if (size > ptp_buffer_put_size)
		{
		  result = ptp_get_type;
		}
	    }
	  else
	    {
	      /* Pack messages */
	      result = ptp_put_type;

	      if (size > ptp_buffer_put_size)
		{
		  result = ptp_get_type;
		}

	      if (msg->message.max_count * msg->message.nb_items >
		  ptp_max_put_nb)
		{
/* 		  result = ptp_copy_put_type; */
		  result = ptp_copy_type;
		}
	    }
	}
    }
  else
    {
      result = ptp_get_type;
    }



#if defined(ELAN_LOG)
#warning "Debug message enable"
  do
    {
      char *t;
      t = sctk_elan_print_message_type (result);
      if (elan_log_file)
	sctk_log (elan_log_file, "size %lu, nb_item %lu, max_count %lu => %s",
		  size, msg->message.nb_items, msg->message.max_count, t);
    }
  while (0);
#endif
  sctk_nodebug ("size %lu, nb_item %lu, max_count %lu => %s",
		size, msg->message.nb_items, msg->message.max_count,
		sctk_elan_print_message_type (result));
  return result;
}

#if 0
static inline int
sctk_elan_lock_manage (int op)
{
  static sctk_spinlock_t lib_elan_big_lock = 0;
  static __thread int owner = 0;

  sctk_nodebug ("owner %d , op %d", owner, op);

  if (op == 0)
    {
      owner--;
      if (owner == 0)
	{
	  sctk_spinlock_unlock (&lib_elan_big_lock);
	}
      return 0;
    }
  if (op == 1)
    {
      sctk_nodebug ("LOCK");
      if (owner >= 1)
	{
	  owner++;
	  return 0;
	}
      else
	{
	  sctk_spinlock_lock (&lib_elan_big_lock);
	  owner = 1;
	  return 0;
	}
      not_reachable ();
    }
  if (op == 2)
    {
      if (owner >= 1)
	{
	  owner++;
	  return 0;
	}
      else
	{
	  if (sctk_spinlock_trylock (&lib_elan_big_lock) == 0)
	    {
	      owner++;
	      return 0;
	    }
	  else
	    {
	      return 1;
	    }
	}
    }
  not_reachable ();
  return -1;
}
#else

static inline int
sctk_elan_lock_manage (int op)
{
  static sctk_thread_mutex_t lib_elan_big_lock =
    SCTK_THREAD_MUTEX_INITIALIZER;
  static __thread int owner = 0;

  sctk_nodebug ("owner %d , op %d", owner, op);

  if (op == 0)
    {
      owner--;
      if (owner == 0)
	{
	  sctk_thread_mutex_unlock (&lib_elan_big_lock);
	}
      return 0;
    }
  if (op == 1)
    {
      sctk_nodebug ("LOCK");
      if (owner >= 1)
	{
	  owner++;
	  return 0;
	}
      else
	{
	  sctk_thread_mutex_lock (&lib_elan_big_lock);
	  owner = 1;
	  return 0;
	}
    }
  if (op == 2)
    {
      if (owner >= 1)
	{
	  owner++;
	  return 0;
	}
      else
	{
	  if (sctk_thread_mutex_trylock (&lib_elan_big_lock) == 0)
	    {
	      owner++;
	      return 0;
	    }
	  else
	    {
	      return 1;
	    }
	}
    }
  not_reachable ();
  return -1;
}
#endif

static inline int
sctk_elan_lock ()
{
  return sctk_elan_lock_manage (1);
}

#if 0
static inline int
sctk_elan_trylock ()
{
  return sctk_elan_lock_manage (2);
}
#endif
static inline int
sctk_elan_unlock ()
{
  return sctk_elan_lock_manage (0);
}

#if defined(ELAN_LOG)
static inline char *
sctk_elan_print_message_type (elan_ptp_message_type_t result)
{
  char *t;

  switch (result)
    {
    case ptp_copy_type:
      {
	t = "ptp_copy_type";
      }
      break;
    case ptp_copy_put_type:
      {
	t = "ptp_copy_put_type";
      }
      break;
    case ptp_put_type:
      {
	t = "ptp_put_type";
      }
      break;
    case ptp_get_type:
      {
	t = "ptp_get_type";
      }
      break;
    default:
      not_reachable ();
    }

  return t;
}
#endif

static __thread volatile int prefered_rail = 0;
static __thread volatile ELAN_PGCTRL *PGCTRL_PTP;
static inline void
sctk_elan_ptp_init_put_get ()
{
  static sctk_spinlock_t lock = 0;
  if (PGCTRL_PTP == NULL)
    {
      sctk_spinlock_lock (&lock);
      if (PGCTRL_PTP == NULL)
	{

	  PGCTRL_PTP = elan_putgetInit (elan_base->state, NULL,
					elan_base->putget_smallputsize,
					elan_base->putget_stripeputsize,
					elan_base->putget_stripegetsize,
					elan_base->putget_throttle,
					elan_base->putget_flags);
	}
      sctk_spinlock_unlock (&lock);
    }
}

typedef struct
{
  volatile size_t size;
  volatile elan_ptp_message_type_t type;
  volatile int nputs;
  volatile long order;
  volatile int rails;
} sctk_elan_ptp_message_t;

static void
sctk_elan_init_sctk_elan_ptp_message_t (sctk_elan_ptp_message_t * m)
{
  m->size = -1;
  m->type = ptp_nothing_type;
  m->nputs = -1;
  m->order = -1;
  m->rails = sctk_rail_number;
}

typedef struct
{
  sctk_elan_ptp_message_t msg;
  sctk_thread_ptp_message_t mpc_msg;
  char buffer[ptp_buffer_size];
  volatile int *completion_flag;
  sctk_atomic_test_t used;
  sctk_spinlock_t spinlock;
  volatile ELAN_EVENT *event;
  volatile ELAN_EVENT *event_rec[SCTK_MAX_RAIL];
  void *addr;
} sctk_elan_ptp_message_buffer_t;

typedef struct
{
  sctk_elan_ptp_message_buffer_t output[ptp_buffer_slots];
  sctk_elan_ptp_message_buffer_t input[ptp_buffer_slots];
  int dest;
  volatile unsigned long in_rank;
  volatile unsigned long out_rank;
  sctk_spinlock_t in_spinlock;
  sctk_spinlock_t out_spinlock;
} sctk_elan_ptp_t;

sctk_elan_ptp_t *sctk_elan_ptp = NULL;

static void
sctk_elan_ptp_init (sctk_elan_ptp_t * slot, int dest)
{
  int i;
  int j;
  sctk_rail_number = elan_nRails (elan_base->state);
  slot->dest = dest;
  slot->in_rank = 0;
  slot->out_rank = 0;
  sctk_spinlock_unlock (&(slot->in_spinlock));
  sctk_spinlock_unlock (&(slot->out_spinlock));
  for (i = 0; i < ptp_buffer_slots; i++)
    {
      sctk_elan_init_sctk_elan_ptp_message_t (&(slot->output[i].msg));
      sctk_elan_init_sctk_elan_ptp_message_t (&(slot->input[i].msg));
      slot->output[i].used = 0;
      slot->input[i].used = 0;
      sctk_spinlock_unlock (&(slot->output[i].spinlock));
      sctk_spinlock_unlock (&(slot->input[i].spinlock));
      slot->output[i].event = NULL;
      slot->input[i].event = NULL;
      memset (&(slot->output[i].mpc_msg), -1,
	      sizeof (sctk_thread_ptp_message_t));
      memset (&(slot->input[i].mpc_msg), -1,
	      sizeof (sctk_thread_ptp_message_t));

      slot->input[i].addr =
	elan_allocElan (elan_base->state, ELAN_EVENT_ALIGN,
			sizeof (ELAN_EVENT_ELAN));
      assume (SCTK_MAX_RAIL >= sctk_rail_number);
      assume (ELAN_RAIL_ALL != sctk_rail_number);
      sctk_nodebug ("ELAN addr %p", slot->input[i].addr);
      for (j = 0; j < sctk_rail_number; j++)
	{
	  sctk_nodebug ("ADDR %d %d %p %p", i, j,
			slot->input[i].addr, slot->input[i].event_rec[j]);
	  slot->input[i].event_rec[j] =
	    elan_initEvent (elan_base->state, elan_base->state->rail[j],
			    (ELAN_ADDR) slot->input[i].addr, 1);
	  sctk_nodebug ("ADDR %d %d %p %p DONE", i, j, slot->input[i].addr,
			slot->input[i].event_rec[j]);
	}
    }
}

static inline void
sctk_elan_init_message_parts (sctk_elan_ptp_message_buffer_t * buf, int nputs,
			      int rails)
{
  int i;
  if (nputs > 1)
    {
      sctk_elan_lock ();
      sctk_nodebug ("Use nputs %d %d REINIT ", nputs, rails);
      if (rails == ELAN_RAIL_ALL)
	{
	  for (i = 0; i < sctk_rail_number; i++)
	    {
	      elan_issueWaitevent ((ELAN_EVENT *) buf->event_rec[i],
				   nputs - 1);
	      /*       _elan_deventDump("",buf->event_rec[i]); */
	    }
	}
      else
	{
	  elan_issueWaitevent ((ELAN_EVENT *) buf->event_rec[rails],
			       nputs - 1);
	  /*       _elan_deventDump("",buf->event_rec[rails]); */
	}
      sctk_elan_unlock ();
    }
}

static void sctk_net_ptp_poll_elan_poll_sends (void);
static inline void
sctk_elan_wait_message_parts (sctk_elan_ptp_message_buffer_t * buf, int rails)
{
  int i;
  sctk_nodebug ("WAIT %d REINIT ", rails);
  if (rails == ELAN_RAIL_ALL)
    {
      for (i = 0; i < sctk_rail_number; i++)
	{
/*       _elan_deventDump("BEFORE",buf->event_rec[i]); */
	  elan_wait ((ELAN_EVENT *) buf->event_rec[i], ELAN_WAIT_EVENT);
/*     _elan_deventDump("AFTER",buf->event_rec[i]); */
	}
    }
  else
    {
/*     _elan_deventDump("BEFORE",buf->event_rec[rails]); */
      elan_wait ((ELAN_EVENT *) buf->event_rec[rails], ELAN_WAIT_EVENT);
/*     _elan_deventDump("AFTER",buf->event_rec[rails]); */
    }
}

static inline size_t
sctk_net_send_ptp_message_driver_item_put (ELAN_EVENT ** event, int slot_nb,
					   sctk_thread_ptp_message_t * msg,
					   int dest_process, int i,
					   sctk_elan_ptp_message_buffer_t *
					   output_msg, size_t pos,
					   ELAN_ADDR addr)
{

  /*Small messages */
  if (msg->message.begins_absolute[i] != NULL)
    {
      /*Absolute addresses */
      int j;
      size_t size;
      sctk_nodebug ("Threat 1 pack with %d items", msg->message.sizes[i]);
      for (j = 0; j < msg->message.sizes[i]; j++)
	{
	  size =
	    (msg->message.ends_absolute[i][j] -
	     msg->message.begins_absolute[i][j] +
	     1) * msg->message.elem_sizes[i];

	  output_msg->msg.nputs++;
	  *event =
	    elan_link (*event,
		       elan_doput ((ELAN_PGCTRL *) PGCTRL_PTP,
				   ((char *) (msg->message.adresses[i])) +
				   msg->message.begins_absolute[i][j] *
				   msg->message.elem_sizes[i],
				   sctk_elan_ptp[sctk_process_rank].
				   input[slot_nb].buffer + pos, addr,
				   size, dest_process,
				   output_msg->msg.rails));
	  pos = pos + size;
	}

      return pos;
    }
  else
    {
      if (msg->message.begins[i] == NULL)
	{
	  size_t size;
	  /*Simple message */
	  sctk_nodebug ("Simple message");
	  size = msg->message.sizes[i];
	  output_msg->msg.nputs++;
	  *event = elan_link (*event, elan_doput ((ELAN_PGCTRL *) PGCTRL_PTP,
						  msg->message.adresses[i],
						  sctk_elan_ptp
						  [sctk_process_rank].
						  input[slot_nb].buffer + pos,
						  addr, size, dest_process,
						  output_msg->msg.rails));
	  return pos + size;
	}
      else
	{
	  int j;
	  size_t size;
	  /*Message pack */
	  sctk_nodebug ("Threat 1 pack with %d items", msg->message.sizes[i]);

	  for (j = 0; j < msg->message.sizes[i]; j++)
	    {
	      size =
		(msg->message.ends[i][j] - msg->message.begins[i][j] + 1) *
		msg->message.elem_sizes[i];

	      output_msg->msg.nputs++;
	      *event =
		elan_link (*event,
			   elan_doput ((ELAN_PGCTRL *) PGCTRL_PTP,
				       (char *) (msg->message.adresses[i]) +
				       msg->message.begins[i][j] *
				       msg->message.elem_sizes[i],
				       sctk_elan_ptp[sctk_process_rank].
				       input[slot_nb].buffer + pos, addr,
				       size, dest_process,
				       output_msg->msg.rails));
	      pos = pos + size;
	    }
	  sctk_nodebug ("Message in buffer|%s|", output_msg->buffer);

	  return pos;
	}
    }
  not_reachable ();
  return 0;
}

static inline size_t
sctk_net_send_ptp_message_driver_item_copy (sctk_thread_ptp_message_t * msg,
					    int i,
					    sctk_elan_ptp_message_buffer_t *
					    output_msg, size_t pos)
{
  /*Small messages */
  if (msg->message.begins_absolute[i] != NULL)
    {
      /*Absolute addresses */
      int j;
      size_t size;
      sctk_nodebug ("Threat 1 pack with %d items", msg->message.sizes[i]);
      for (j = 0; j < msg->message.sizes[i]; j++)
	{
	  size =
	    (msg->message.ends_absolute[i][j] -
	     msg->message.begins_absolute[i][j] +
	     1) * msg->message.elem_sizes[i];

	  memcpy (output_msg->buffer + pos,
		  ((char *) (msg->message.adresses[i])) +
		  msg->message.begins_absolute[i][j] *
		  msg->message.elem_sizes[i], size);
	  pos = pos + size;
	}

      return pos;
    }
  else
    {
      if (msg->message.begins[i] == NULL)
	{
	  size_t size;
	  /*Simple message */
	  sctk_nodebug ("Simple message");
	  size = msg->message.sizes[i];
	  memcpy (output_msg->buffer + pos, msg->message.adresses[i], size);
	  return pos + size;
	}
      else
	{
	  int j;
	  size_t size;
	  /*Message pack */
	  sctk_nodebug ("Threat 1 pack with %d items", msg->message.sizes[i]);

	  for (j = 0; j < msg->message.sizes[i]; j++)
	    {
	      size =
		(msg->message.ends[i][j] - msg->message.begins[i][j] + 1) *
		msg->message.elem_sizes[i];

	      memcpy (output_msg->buffer + pos,
		      (char *) (msg->message.adresses[i]) +
		      msg->message.begins[i][j] * msg->message.elem_sizes[i],
		      size);
	      pos = pos + size;
	    }
	  sctk_nodebug ("Message in buffer|%s|", output_msg->buffer);

	  return pos;
	}
    }
  not_reachable ();
  return 0;
}

static inline size_t
sctk_net_send_ptp_message_driver_item_copy_put (ELAN_EVENT ** event,
						int slot_nb,
						sctk_thread_ptp_message_t *
						msg, int dest_process, int i,
						sctk_elan_ptp_message_buffer_t
						* output_msg, size_t pos,
						size_t * last_put,
						ELAN_ADDR addr)
{
  size_t new_pos;
  size_t last;

  last = *last_put;
  new_pos = last + ptp_buffer_put_size;
  if (pos >= new_pos)
    {
      output_msg->msg.nputs++;
      *event = elan_link (*event, elan_doput ((ELAN_PGCTRL *) PGCTRL_PTP,
					      output_msg->buffer + last,
					      sctk_elan_ptp
					      [sctk_process_rank].
					      input[slot_nb].buffer + last,
					      addr, pos - last, dest_process,
					      output_msg->msg.rails));
      *last_put = pos;
    }

  /*Small messages */
  if (msg->message.begins_absolute[i] != NULL)
    {
      /*Absolute addresses */
      int j;
      size_t size;
      sctk_nodebug ("Threat 1 pack with %d items", msg->message.sizes[i]);
      for (j = 0; j < msg->message.sizes[i]; j++)
	{
	  size =
	    (msg->message.ends_absolute[i][j] -
	     msg->message.begins_absolute[i][j] +
	     1) * msg->message.elem_sizes[i];

	  memcpy (output_msg->buffer + pos,
		  ((char *) (msg->message.adresses[i])) +
		  msg->message.begins_absolute[i][j] *
		  msg->message.elem_sizes[i], size);
	  pos = pos + size;
	}

      return pos;
    }
  else
    {
      if (msg->message.begins[i] == NULL)
	{
	  size_t size;
	  /*Simple message */
	  sctk_nodebug ("Simple message");
	  size = msg->message.sizes[i];
	  memcpy (output_msg->buffer + pos, msg->message.adresses[i], size);
	  return pos + size;
	}
      else
	{
	  int j;
	  size_t size;
	  /*Message pack */
	  sctk_nodebug ("Threat 1 pack with %d items", msg->message.sizes[i]);

	  for (j = 0; j < msg->message.sizes[i]; j++)
	    {
	      size =
		(msg->message.ends[i][j] - msg->message.begins[i][j] + 1) *
		msg->message.elem_sizes[i];

	      memcpy (output_msg->buffer + pos,
		      (char *) (msg->message.adresses[i]) +
		      msg->message.begins[i][j] * msg->message.elem_sizes[i],
		      size);
	      pos = pos + size;
	    }
	  sctk_nodebug ("Message in buffer|%s|", output_msg->buffer);

	  return pos;
	}
    }
  not_reachable ();
  return 0;
}

static inline void
sctk_net_send_ptp_message_driver_flush (sctk_elan_ptp_t * current, int i)
{
  if (current->output[i].event != NULL)
    {
      sctk_nodebug ("Try to flush %d", i);
      if (elan_poll ((ELAN_EVENT *) current->output[i].event, 0))
	{
	  if (current->output[i].completion_flag)
	    {
	      sctk_nodebug ("release %d", i);
	      *(current->output[i].completion_flag) = 1;
	      current->output[i].completion_flag = NULL;
	    }
	  current->output[i].event = NULL;
	}
    }
}

#define sctk_net_send_output_start msg

static void
sctk_net_send_ptp_message_driver (sctk_thread_ptp_message_t * msg,
				  int dest_process)
{
  sctk_elan_ptp_t *current;
  int i;
  size_t total_size = 0;
  sctk_elan_ptp_message_buffer_t *output_msg;
  size_t pos = 0;
  int slot_nb;
  int msg_rank;
  ELAN_EVENT *event = NULL;
  elan_ptp_message_type_t msg_type;
  int *compl_flag;
  SCTK_PROFIL_START (sctk_net_send_ptp_message_driver);
  sctk_nodebug ("in");
  sctk_elan_ptp_init_put_get ();

  prefered_rail = (prefered_rail + 1) % sctk_rail_number;

  current = &(sctk_elan_ptp[dest_process]);

  for (i = 0; i < msg->message.nb_items; i++)
    {
      if (msg->message.begins_absolute[i] != NULL)
	{
	  int j;
	  for (j = 0; j < msg->message.sizes[i]; j++)
	    {
	      total_size +=
		(msg->message.ends_absolute[i][j] -
		 msg->message.begins_absolute[i][j] +
		 1) * msg->message.elem_sizes[i];
	    }
	}
      else
	{
	  if (msg->message.begins[i] == NULL)
	    {
	      total_size += msg->message.sizes[i];
	    }
	  else
	    {
	      int j;
	      for (j = 0; j < msg->message.sizes[i]; j++)
		{
		  total_size +=
		    (msg->message.ends[i][j] - msg->message.begins[i][j] +
		     1) * msg->message.elem_sizes[i];
		}
	    }
	}
    }

  sctk_nodebug ("TRY NEW SEND");
  sctk_nodebug ("NEW SEND");

  while (1)
    {
      for (i = 0; i < ptp_buffer_slots; i++)
	{
	  if (current->output[i].used == 0)
	    {
	      if (sctk_test_and_set (&(current->output[i].used)) == 0)
		{
		  slot_nb = i;
		  sctk_spinlock_lock (&(current->output[slot_nb].spinlock));

		  if (current->output[i].event != NULL)
		    {
		      elan_wait ((ELAN_EVENT *) current->output[i].event,
				 ELAN_WAIT_EVENT);
		      current->output[i].event = NULL;
		    }

		  if (current->output[i].completion_flag)
		    {
		      *(current->output[i].completion_flag) = 1;
		      current->output[i].completion_flag = NULL;
		      sctk_nodebug ("release %d", i);
		    }

		  goto slot_found;
		}
	    }
	}
      sctk_net_ptp_poll_elan (NULL);
      sctk_thread_yield ();
    }

slot_found:
  sctk_nodebug ("out 0");
/*   while(sctk_spinlock_trylock(&(current->out_spinlock)) != 0); */
  sctk_spinlock_lock (&(current->out_spinlock));
  msg_rank = current->out_rank;
  current->out_rank++;
  sctk_spinlock_unlock (&(current->out_spinlock));

  assume (slot_nb < ptp_buffer_slots);
  assume (slot_nb >= 0);

  output_msg = &(current->output[slot_nb]);

  sctk_nodebug ("SEND Use slot %d for %d %d %lu", slot_nb, dest_process,
		msg_rank, total_size);
  sctk_nodebug ("Total size %lu", total_size);

  sctk_elan_init_sctk_elan_ptp_message_t (&(output_msg->msg));
  msg->completion_flag = 0;
  compl_flag = (int *) &(msg->completion_flag);
  output_msg->msg.size = total_size;

  output_msg->msg.order = msg_rank;
  msg->process_src = sctk_process_rank;
  msg->addr_to_free = (void *) &(msg->completion_flag);

  sctk_nodebug ("SEND message from % to %d size %lu",
		msg->header.source, msg->header.destination, total_size);

  msg_type = sctk_elan_determine_message_type (total_size, msg);
  output_msg->msg.type = msg_type;
  output_msg->msg.nputs = 1;
  output_msg->msg.rails = ELAN_RAIL_ALL;

  memcpy (&(output_msg->mpc_msg), msg, sizeof (sctk_thread_ptp_message_t));


  sctk_nodebug ("out 1");
  switch (msg_type)
    {
    case ptp_copy_type:
      {
	size_t local_size;
	msg->addr_to_free = NULL;
	for (i = 0; i < msg->message.nb_items; i++)
	  {
	    pos = sctk_net_send_ptp_message_driver_item_copy (msg, i,
							      output_msg,
							      pos);
	    sctk_nodebug ("New pos %lu", pos);
	    assume (pos <= total_size);
	  }
	assume (pos == total_size);
	sctk_nodebug ("SEND Message in buffer|%s|", output_msg->buffer);

	local_size = (unsigned long) (output_msg->buffer + total_size) -
	  (unsigned long) (&(output_msg->sctk_net_send_output_start));

	sctk_nodebug
	  ("msg_size %lu local_size %lu %lu+%lu+%lu header %lu slot %d rank %d heder_msg_mpc %lu",
	   total_size, local_size, total_size,
	   sizeof (sctk_thread_ptp_message_t),
	   sizeof (sctk_elan_ptp_message_t),
	   sizeof (sctk_thread_ptp_message_t), slot_nb, msg_rank,
	   sizeof (sctk_thread_message_header_t));

	sctk_nodebug ("SEND ptp_copy_type nputs %d %d", output_msg->msg.nputs,
		      slot_nb);

	output_msg->msg.rails = prefered_rail;
	event =
	  elan_doput ((ELAN_PGCTRL *) PGCTRL_PTP,
		      &(output_msg->sctk_net_send_output_start),
		      &(sctk_elan_ptp[sctk_process_rank].input[slot_nb].
			sctk_net_send_output_start),
		      (ELAN_ADDR) sctk_elan_ptp[sctk_process_rank].
		      input[slot_nb].addr, local_size, dest_process,
		      output_msg->msg.rails);

	if (local_size <= elan_base->putget_smallputsize)
	  {
	    elan_wait (event, ELAN_WAIT_EVENT);
	    event = NULL;
	  }
	if (compl_flag)
	  {
	    sctk_nodebug ("release %d", slot_nb);
	    *(compl_flag) = 1;
	  }
	compl_flag = NULL;
      }
      break;
    case ptp_copy_put_type:
      {
	size_t local_size;
	size_t last_put = 0;
	msg->addr_to_free = NULL;
	sctk_nodebug ("Entering COPY PUT");
	for (i = 0; i < msg->message.nb_items; i++)
	  {
	    pos =
	      sctk_net_send_ptp_message_driver_item_copy_put (&event, slot_nb,
							      msg,
							      dest_process, i,
							      output_msg, pos,
							      &last_put,
							      (ELAN_ADDR)
							      sctk_elan_ptp
							      [sctk_process_rank].
							      input[slot_nb].
							      addr);
	    sctk_nodebug ("New pos %lu", pos);
	    assume (pos <= total_size);
	  }
	assume (pos == total_size);
	sctk_nodebug ("SEND Message in buffer|%s|", output_msg->buffer);

	if (pos - last_put > 0)
	  {
	    output_msg->msg.nputs++;
	    event = elan_link (event, elan_doput ((ELAN_PGCTRL *) PGCTRL_PTP,
						  output_msg->buffer +
						  last_put,
						  sctk_elan_ptp
						  [sctk_process_rank].
						  input[slot_nb].buffer +
						  last_put,
						  (ELAN_ADDR)
						  sctk_elan_ptp
						  [sctk_process_rank].
						  input[slot_nb].addr,
						  pos - last_put,
						  dest_process,
						  output_msg->msg.rails));
	  }

	sctk_nodebug ("msg_size %lu header %lu",
		      total_size, sizeof (sctk_thread_ptp_message_t));

	local_size = (unsigned long) (output_msg->buffer) -
	  (unsigned long) (&(output_msg->sctk_net_send_output_start));

	sctk_nodebug ("SEND ptp_copy_put_type nputs %d %d",
		      output_msg->msg.nputs, slot_nb);
	event =
	  elan_link (event,
		     elan_doput ((ELAN_PGCTRL *) PGCTRL_PTP,
				 &(output_msg->sctk_net_send_output_start),
				 &(sctk_elan_ptp[sctk_process_rank].
				   input[slot_nb].sctk_net_send_output_start),
				 (ELAN_ADDR) sctk_elan_ptp[sctk_process_rank].
				 input[slot_nb].addr, local_size,
				 dest_process, output_msg->msg.rails));
	if (compl_flag)
	  {
	    sctk_nodebug ("release %d", slot_nb);
	    *(compl_flag) = 1;
	  }
	compl_flag = NULL;
      }
      break;
    case ptp_put_type:
      {
	size_t local_size;
	msg->addr_to_free = NULL;
	sctk_nodebug ("Entering PUT");
	for (i = 0; i < msg->message.nb_items; i++)
	  {
	    pos =
	      sctk_net_send_ptp_message_driver_item_put (&event, slot_nb, msg,
							 dest_process, i,
							 output_msg, pos,
							 (ELAN_ADDR)
							 sctk_elan_ptp
							 [sctk_process_rank].
							 input[slot_nb].addr);
	    assume (pos <= total_size);
	    sctk_nodebug ("New pos %lu", pos);
	  }
	assume (pos == total_size);

	local_size = (unsigned long) (output_msg->buffer) -
	  (unsigned long) (&(output_msg->sctk_net_send_output_start));

	sctk_nodebug ("SEND ptp_put_type nputs %d slot %d %p",
		      output_msg->msg.nputs, slot_nb,
		      sctk_elan_ptp[sctk_process_rank].input[slot_nb].addr);
	event =
	  elan_link (event,
		     elan_doput ((ELAN_PGCTRL *) PGCTRL_PTP,
				 &(output_msg->sctk_net_send_output_start),
				 &(sctk_elan_ptp[sctk_process_rank].
				   input[slot_nb].sctk_net_send_output_start),
				 (ELAN_ADDR) sctk_elan_ptp[sctk_process_rank].
				 input[slot_nb].addr, local_size,
				 dest_process, output_msg->msg.rails));
	sctk_nodebug ("Leaving PUT");
      }
      break;
    case ptp_get_type:
      {
	size_t local_size;

	compl_flag = NULL;

	local_size = (unsigned long) (output_msg->buffer) -
	  (unsigned long) (&(output_msg->sctk_net_send_output_start));

	sctk_nodebug ("SEND ptp_get_type nputs %d %d", output_msg->msg.nputs,
		      slot_nb);
	output_msg->msg.rails = prefered_rail;
	event =
	  elan_doput ((ELAN_PGCTRL *) PGCTRL_PTP,
		      &(output_msg->sctk_net_send_output_start),
		      &(sctk_elan_ptp[sctk_process_rank].input[slot_nb].
			sctk_net_send_output_start),
		      (ELAN_ADDR) sctk_elan_ptp[sctk_process_rank].
		      input[slot_nb].addr, local_size, dest_process,
		      output_msg->msg.rails);

	elan_wait (event, ELAN_WAIT_EVENT);
	event = NULL;
      }
      break;
    default:
      not_reachable ();
    }

  sctk_nodebug ("out 2");
  if (event != NULL)
    {
      if (elan_poll (event, 0))
	{
	  if (compl_flag)
	    {
	      sctk_nodebug ("release %d", slot_nb);
	      *(compl_flag) = 1;
	    }
	  current->output[slot_nb].event = NULL;
	}
      else
	{
	  current->output[slot_nb].completion_flag = compl_flag;
	  current->output[slot_nb].event = event;
	}
    }
  else
    {
      if (compl_flag)
	{
	  sctk_nodebug ("release %d", slot_nb);
	  *(compl_flag) = 1;
	}
      current->output[slot_nb].event = NULL;
    }

  sctk_spinlock_unlock (&(current->output[slot_nb].spinlock));
  sctk_nodebug ("SEND DONE");
  sctk_nodebug ("out");
  SCTK_PROFIL_END (sctk_net_send_ptp_message_driver);
}

static inline void
sctk_net_ptp_message_done (sctk_thread_ptp_message_t * item)
{
  sctk_free (item);
}

static inline int
sctk_net_ptp_poll_elan_one_process (ELAN_EVENT ** event_g, int i)
{
  sctk_elan_ptp_t *current;
  sctk_thread_ptp_message_t *msg;
  ELAN_EVENT *event = NULL;
  long cur_rank;
  int j;
  static sctk_atomic_test_t used = 0;
  int pending = 0;

  current = &(sctk_elan_ptp[i]);

  for (j = 0; j < ptp_buffer_slots; j++)
    {
      cur_rank = current->in_rank;
      if ((current->input[j].msg.type != ptp_nothing_type)
	  && (current->input[j].msg.size != -1)
	  && (current->input[j].msg.nputs != -1)
	  && (current->input[j].msg.order == cur_rank)
	  && (current->input[j].msg.rails != sctk_rail_number))
	{
	  size_t size;

	  void *tmp;
	  char *buffer;

	  if (sctk_spinlock_trylock (&(current->input[j].spinlock)) == 0)
	    {
	      if ((current->input[j].msg.type != ptp_nothing_type)
		  && (current->input[j].msg.size != -1)
		  && (current->input[j].msg.nputs != -1)
		  && (current->input[j].msg.order == cur_rank)
		  && (current->input[j].msg.rails != sctk_rail_number))
		{
		  int nputs;
		  int rails;
		  rails = current->input[j].msg.rails;
		  size = current->input[j].msg.size;
		  nputs = current->input[j].msg.nputs;
		  current->input[j].msg.size = -1;
		  current->input[j].msg.nputs = -1;

		  sctk_nodebug ("%d nputs", nputs);
		  sctk_elan_init_message_parts (&(current->input[j]), nputs,
						rails);

		  switch (current->input[j].msg.type)
		    {
		    case ptp_copy_type:
		    case ptp_copy_put_type:
		    case ptp_put_type:
		      {
			tmp =
			  sctk_malloc (sizeof (sctk_thread_ptp_message_t) +
				       size);

			sctk_elan_wait_message_parts (&(current->input[j]),
						      rails);

			memcpy (tmp, &(current->input[j].mpc_msg),
				sizeof (sctk_thread_ptp_message_t) + size);

			msg = (sctk_thread_ptp_message_t *) tmp;
			buffer =
			  ((char *) tmp + sizeof (sctk_thread_ptp_message_t));
		      }
		      break;
		    case ptp_get_type:
		      {
			tmp =
			  sctk_malloc (sctk_aligned_sizeof
				       (sctk_thread_ptp_message_t));

			sctk_elan_wait_message_parts (&(current->input[j]),
						      rails);

			memcpy (tmp, &(current->input[j].mpc_msg),
				sizeof (sctk_thread_ptp_message_t));
			msg = (sctk_thread_ptp_message_t *) tmp;
			buffer = (void *) -1;
		      }
		      break;
		    default:
		      not_reachable ();
		    }

		  msg->net_mesg = buffer;
		  sctk_nodebug ("alloc %p", tmp);

/* #warning "For debug purposes only" */
/* 		  memset (&(current->input[j].mpc_msg), -1, */
/* 			  sizeof (sctk_thread_ptp_message_t)); */

		  sctk_elan_init_sctk_elan_ptp_message_t (&
							  (current->input[j].
							   msg));

		  sctk_send_message (msg);

		  sctk_spinlock_lock (&(current->in_spinlock));
		  current->in_rank++;
		  sctk_spinlock_unlock (&(current->in_spinlock));

		  event =
		    elan_link (event,
			       elan_doput ((ELAN_PGCTRL *) PGCTRL_PTP,
					   (void *) &used,
					   (void *)
					   &(sctk_elan_ptp[sctk_process_rank].
					     output[j].used),
					   (ELAN_ADDR) NULL,
					   sizeof (sctk_atomic_test_t), i,
					   prefered_rail));
		}
	      sctk_spinlock_unlock (&(current->input[j].spinlock));
	    }
	}
      else
	{
	  if ((current->input[j].msg.type != ptp_nothing_type)
	      && (current->input[j].msg.size != -1)
	      && (current->input[j].msg.nputs != -1)
	      && (current->input[j].msg.rails != sctk_rail_number))
	    {
	      pending++;
	    }
	}
    }

  if (event)
    {
      *event_g = elan_link (*event_g, event);
    }
  return pending;
}

static void
sctk_net_ptp_poll_elan_poll_sends ()
{
  int i;
  int j;
  sctk_elan_ptp_t *current;
  for (i = 0; i < sctk_process_number; i++)
    {
      current = &(sctk_elan_ptp[i]);
      for (j = 0; j < ptp_buffer_slots; j++)
	{
	  if (current->output[j].event != NULL)
	    {
	      if (sctk_spinlock_trylock (&(current->output[j].spinlock)) == 0)
		{
		  sctk_net_send_ptp_message_driver_flush (current, j);
		  sctk_spinlock_unlock (&(current->output[j].spinlock));
		}
	    }
	}
    }
}

static int
sctk_net_ptp_poll_elan_intern (void *arg, int mode)
{
  int i;
  int res = 0;
  static __thread ELAN_EVENT *event = NULL;
  static sctk_spinlock_t lock = 0;

  sctk_elan_ptp_init_put_get ();

  sctk_net_ptp_poll_elan_poll_sends ();

  sctk_nodebug ("NEW POLL");
  /*Message polling */
  for (i = 0; i < sctk_process_number; i++)
    {
      res += sctk_net_ptp_poll_elan_one_process (&event, i);
    }

  sctk_nodebug ("END POLL");

  if (mode)
    {
      if (event)
	{
	  elan_wait (event, ELAN_WAIT_EVENT);
	  event = NULL;
	}
    }
  else
    {
      if (event)
	{
	  if (elan_poll (event, 0))
	    {
	      event = NULL;
	    }

	}
    }
  return res;
}
static int
sctk_net_ptp_poll_elan_thread (void *arg)
{
  return sctk_net_ptp_poll_elan_intern (arg, 0);
}
static void
sctk_net_ptp_poll_elan (void *arg)
{
  sctk_net_ptp_poll_elan_intern (arg, 1);
}

static void *
sctk_elan_poll_thread (void *arg)
{
  while (1)
    {
      int res;
      sctk_net_adm_poll_elan (NULL);
      res = sctk_net_ptp_poll_elan_thread (NULL);
      if (!res)
	usleep (1);
      else
	sched_yield ();
    }
}

static void
sctk_launch_poll_thread ()
{
  sctk_thread_t pid;
  sctk_thread_attr_t attr;
  sctk_thread_attr_init (&attr);
  sctk_thread_attr_setscope (&attr, SCTK_THREAD_SCOPE_SYSTEM);

  sctk_user_thread_create (&pid, &attr, sctk_elan_poll_thread, NULL);
}

#define require_malloc_ptp(a) require+= sctk_aligned_size((a))
#define get_malloc_ptp(b,a) b = (void*)current;current += sctk_aligned_size((a))

static inline ELAN_EVENT *
remote_memcpy_intern (void *src, void *dest, size_t size,
		      int dest_p, ELAN_EVENT ** event)
{
  static __thread int pending = 0;
  if (*event)
    {
      if (pending >= 10)
	{
	  pending = 0;
	  elan_wait (*event, ELAN_WAIT_EVENT);
	  *event = NULL;
	}
    }

  *event = elan_link (*event, elan_doget ((ELAN_PGCTRL *) PGCTRL_PTP,
					  src,
					  dest, size, dest_p, ELAN_RAIL_ALL));
  pending++;
  return *event;

}

#define remote_memcpy(a,b,c) *event = remote_memcpy_intern(b,a,c,dest,event)

static inline void
sctk_elan_copy_buffer_std_std (ELAN_EVENT ** event,
			       sctk_pack_indexes_t * restrict in_begins,
			       sctk_pack_indexes_t * restrict in_ends,
			       size_t in_sizes,
			       void *restrict in_adress,
			       size_t in_elem_size,
			       sctk_pack_indexes_t * restrict out_begins,
			       sctk_pack_indexes_t * restrict out_ends,
			       size_t out_sizes,
			       void *restrict out_adress,
			       size_t out_elem_size, int dest)
{
  sctk_pack_indexes_t tmp_begin[1];
  sctk_pack_indexes_t tmp_end[1];
  if ((in_begins == NULL) && (out_begins == NULL))
    {
      sctk_nodebug ("%s == %s", out_adress, in_adress);
      remote_memcpy (out_adress, in_adress, in_sizes);
      sctk_nodebug ("%s == %s", out_adress, in_adress);
    }
  else
    {
      unsigned long i;
      unsigned long j;
      unsigned long in_i;
      unsigned long in_j;
      if (in_begins == NULL)
	{
	  in_begins = tmp_begin;
	  in_begins[0] = 0;
	  in_ends = tmp_end;
	  in_ends[0] = in_sizes - 1;
	  in_elem_size = 1;
	  in_sizes = 1;
	}
      if (out_begins == NULL)
	{
	  out_begins = tmp_begin;
	  out_begins[0] = 0;
	  out_ends = tmp_end;
	  out_ends[0] = out_sizes - 1;
	  out_elem_size = 1;
	  out_sizes = 1;
	}
      in_i = 0;
      in_j = in_begins[in_i] * in_elem_size;
      for (i = 0; i < out_sizes; i++)
	{
	  for (j = out_begins[i] * out_elem_size;
	       j <= out_ends[i] * out_elem_size;)
	    {
	      size_t max_length;
	      if (in_j > in_ends[in_i] * in_elem_size)
		{
		  in_i++;
		  if (in_i > in_sizes)
		    {
		      return;
		    }
		  in_j = in_begins[in_i] * in_elem_size;
		}

	      max_length =
		sctk_min ((out_ends[i] * out_elem_size - j +
			   out_elem_size),
			  (in_ends[in_i] * in_elem_size - in_j +
			   in_elem_size));

	      remote_memcpy ((((char *) out_adress) + j),
			     (((char *) in_adress) + in_j), max_length);
	      sctk_nodebug ("Copy out[%d-%d]%s == in[%d-%d]%s", j,
			    j + max_length, (((char *) out_adress) + j),
			    in_j, in_j + max_length,
			    (((char *) in_adress) + in_j));

	      j += max_length;
	      in_j += max_length;

	    }
	}
    }
}

static inline void
sctk_elan_copy_buffer_absolute_absolute (ELAN_EVENT ** event,
					 sctk_pack_absolute_indexes_t *
					 restrict in_begins,
					 sctk_pack_absolute_indexes_t *
					 restrict in_ends, size_t in_sizes,
					 void *restrict in_adress,
					 size_t in_elem_size,
					 sctk_pack_absolute_indexes_t *
					 restrict out_begins,
					 sctk_pack_absolute_indexes_t *
					 restrict out_ends, size_t out_sizes,
					 void *restrict out_adress,
					 size_t out_elem_size, int dest)
{
  sctk_pack_absolute_indexes_t tmp_begin[1];
  sctk_pack_absolute_indexes_t tmp_end[1];
  if ((in_begins == NULL) && (out_begins == NULL))
    {
      sctk_nodebug ("%s == %s", out_adress, in_adress);
      remote_memcpy (out_adress, in_adress, in_sizes);
      sctk_nodebug ("%s == %s", out_adress, in_adress);
    }
  else
    {
      unsigned long i;
      unsigned long j;
      unsigned long in_i;
      unsigned long in_j;
      if (in_begins == NULL)
	{
	  in_begins = tmp_begin;
	  in_begins[0] = 0;
	  in_ends = tmp_end;
	  in_ends[0] = in_sizes - 1;
	  in_elem_size = 1;
	  in_sizes = 1;
	}
      if (out_begins == NULL)
	{
	  out_begins = tmp_begin;
	  out_begins[0] = 0;
	  out_ends = tmp_end;
	  out_ends[0] = out_sizes - 1;
	  out_elem_size = 1;
	  out_sizes = 1;
	}
      in_i = 0;
      in_j = in_begins[in_i] * in_elem_size;
      for (i = 0; i < out_sizes; i++)
	{
	  for (j = out_begins[i] * out_elem_size;
	       j <= out_ends[i] * out_elem_size;)
	    {
	      size_t max_length;
	      if (in_j > in_ends[in_i] * in_elem_size)
		{
		  in_i++;
		  if (in_i > in_sizes)
		    {
		      return;
		    }
		  in_j = in_begins[in_i] * in_elem_size;
		}

	      max_length =
		sctk_min ((out_ends[i] * out_elem_size - j +
			   out_elem_size),
			  (in_ends[in_i] * in_elem_size - in_j +
			   in_elem_size));

	      sctk_nodebug ("Copy out[%lu-%lu]%p == in[%lu-%lu]%p", j,
			    j + max_length, &(((char *) out_adress)[j]),
			    in_j, in_j + max_length,
			    &(((char *) in_adress)[in_j]));
	      remote_memcpy ((((char *) out_adress) + j),
			     (((char *) in_adress) + in_j), max_length);
	      sctk_nodebug ("Copy out[%d-%d]%s == in[%d-%d]%s", j,
			    j + max_length, &(((char *) out_adress)[j]),
			    in_j, in_j + max_length,
			    &(((char *) in_adress)[in_j]));

	      j += max_length;
	      in_j += max_length;
	    }
	}
    }
}

static inline void
sctk_elan_copy_buffer_absolute_std (ELAN_EVENT ** event,
				    sctk_pack_absolute_indexes_t *
				    restrict in_begins,
				    sctk_pack_absolute_indexes_t *
				    restrict in_ends, size_t in_sizes,
				    void *restrict in_adress,
				    size_t in_elem_size,
				    sctk_pack_indexes_t * restrict out_begins,
				    sctk_pack_indexes_t * restrict out_ends,
				    size_t out_sizes,
				    void *restrict out_adress,
				    size_t out_elem_size, int dest)
{
  sctk_pack_indexes_t tmp_begin[1];
  sctk_pack_indexes_t tmp_end[1];
  sctk_pack_absolute_indexes_t tmp_begin_absolute[1];
  sctk_pack_absolute_indexes_t tmp_end_absolute[1];
  if ((in_begins == NULL) && (out_begins == NULL))
    {
      sctk_nodebug ("%s == %s", out_adress, in_adress);
      remote_memcpy (out_adress, in_adress, in_sizes);
      sctk_nodebug ("%s == %s", out_adress, in_adress);
    }
  else
    {
      unsigned long i;
      unsigned long j;
      unsigned long in_i;
      unsigned long in_j;
      if (in_begins == NULL)
	{
	  in_begins = tmp_begin_absolute;
	  in_begins[0] = 0;
	  in_ends = tmp_end_absolute;
	  in_ends[0] = in_sizes - 1;
	  in_elem_size = 1;
	  in_sizes = 1;
	}
      if (out_begins == NULL)
	{
	  out_begins = tmp_begin;
	  out_begins[0] = 0;
	  out_ends = tmp_end;
	  out_ends[0] = out_sizes - 1;
	  out_elem_size = 1;
	  out_sizes = 1;
	}
      in_i = 0;
      in_j = in_begins[in_i] * in_elem_size;
      for (i = 0; i < out_sizes; i++)
	{
	  for (j = out_begins[i] * out_elem_size;
	       j <= out_ends[i] * out_elem_size;)
	    {
	      size_t max_length;
	      if (in_j > in_ends[in_i] * in_elem_size)
		{
		  in_i++;
		  if (in_i > in_sizes)
		    {
		      return;
		    }
		  in_j = in_begins[in_i] * in_elem_size;
		}

	      max_length =
		sctk_min ((out_ends[i] * out_elem_size - j +
			   out_elem_size),
			  (in_ends[in_i] * in_elem_size - in_j +
			   in_elem_size));

	      remote_memcpy ((((char *) out_adress) + j),
			     (((char *) in_adress) + in_j), max_length);
	      sctk_nodebug ("Copy out[%d-%d]%s == in[%d-%d]%s", j,
			    j + max_length, &(((char *) out_adress)[j]),
			    in_j, in_j + max_length,
			    &(((char *) in_adress)[in_j]));

	      j += max_length;
	      in_j += max_length;
	    }
	}
    }
}

static inline void
sctk_elan_copy_buffer_std_absolute (ELAN_EVENT ** event,
				    sctk_pack_indexes_t * restrict in_begins,
				    sctk_pack_indexes_t * restrict in_ends,
				    size_t in_sizes, void *restrict in_adress,
				    size_t in_elem_size,
				    sctk_pack_absolute_indexes_t *
				    restrict out_begins,
				    sctk_pack_absolute_indexes_t *
				    restrict out_ends, size_t out_sizes,
				    void *restrict out_adress,
				    size_t out_elem_size, int dest)
{
  sctk_pack_indexes_t tmp_begin[1];
  sctk_pack_indexes_t tmp_end[1];
  sctk_pack_absolute_indexes_t tmp_begin_absolute[1];
  sctk_pack_absolute_indexes_t tmp_end_absolute[1];
  if ((in_begins == NULL) && (out_begins == NULL))
    {
      sctk_nodebug ("%s == %s", out_adress, in_adress);
      remote_memcpy (out_adress, in_adress, in_sizes);
      sctk_nodebug ("%s == %s", out_adress, in_adress);
    }
  else
    {
      unsigned long i;
      unsigned long j;
      unsigned long in_i;
      unsigned long in_j;
      if (in_begins == NULL)
	{
	  in_begins = tmp_begin;
	  in_begins[0] = 0;
	  in_ends = tmp_end;
	  in_ends[0] = in_sizes - 1;
	  in_elem_size = 1;
	  in_sizes = 1;
	}
      if (out_begins == NULL)
	{
	  out_begins = tmp_begin_absolute;
	  out_begins[0] = 0;
	  out_ends = tmp_end_absolute;
	  out_ends[0] = out_sizes - 1;
	  out_elem_size = 1;
	  out_sizes = 1;
	}
      in_i = 0;
      in_j = in_begins[in_i] * in_elem_size;
      for (i = 0; i < out_sizes; i++)
	{
	  for (j = out_begins[i] * out_elem_size;
	       j <= out_ends[i] * out_elem_size;)
	    {
	      size_t max_length;
	      if (in_j > in_ends[in_i] * in_elem_size)
		{
		  in_i++;
		  if (in_i > in_sizes)
		    {
		      return;
		    }
		  in_j = in_begins[in_i] * in_elem_size;
		}

	      max_length =
		sctk_min ((out_ends[i] * out_elem_size - j +
			   out_elem_size),
			  (in_ends[in_i] * in_elem_size - in_j +
			   in_elem_size));

	      remote_memcpy ((((char *) out_adress) + j),
			     (((char *) in_adress) + in_j), max_length);
	      sctk_nodebug ("Copy out[%d-%d]%s == in[%d-%d]%s", j,
			    j + max_length, &(((char *) out_adress)[j]),
			    in_j, in_j + max_length,
			    &(((char *) in_adress)[in_j]));

	      j += max_length;
	      in_j += max_length;
	    }
	}
    }
}

static inline void
sctk_elan_copy_buffer (ELAN_EVENT ** event,
		       sctk_pack_indexes_t * restrict in_begins,
		       sctk_pack_indexes_t * restrict in_ends,
		       sctk_pack_absolute_indexes_t *
		       restrict in_begins_absolute,
		       sctk_pack_absolute_indexes_t *
		       restrict in_ends_absolute, size_t in_sizes,
		       void *restrict in_adress, size_t in_elem_size,
		       sctk_pack_indexes_t * restrict out_begins,
		       sctk_pack_indexes_t * restrict out_ends,
		       sctk_pack_absolute_indexes_t *
		       restrict out_begins_absolute,
		       sctk_pack_absolute_indexes_t *
		       restrict out_ends_absolute, size_t out_sizes,
		       void *restrict out_adress, size_t out_elem_size,
		       int dest)
{
  sctk_nodebug ("in %p out %p", in_adress, out_adress);
  if ((in_begins_absolute == NULL) && (out_begins_absolute == NULL))
    {
      sctk_elan_copy_buffer_std_std (event,
				     in_begins,
				     in_ends,
				     in_sizes,
				     in_adress,
				     in_elem_size,
				     out_begins,
				     out_ends,
				     out_sizes, out_adress, out_elem_size,
				     dest);
    }
  else if ((in_begins_absolute != NULL) && (out_begins_absolute != NULL))
    {
      sctk_elan_copy_buffer_absolute_absolute (event, in_begins_absolute,
					       in_ends_absolute, in_sizes,
					       in_adress, in_elem_size,
					       out_begins_absolute,
					       out_ends_absolute, out_sizes,
					       out_adress, out_elem_size,
					       dest);
    }
  else if ((in_begins_absolute == NULL) && (out_begins_absolute != NULL))
    {
      sctk_elan_copy_buffer_std_absolute (event, in_begins, in_ends, in_sizes,
					  in_adress, in_elem_size,
					  out_begins_absolute,
					  out_ends_absolute, out_sizes,
					  out_adress, out_elem_size, dest);
    }
  else
    {
      sctk_elan_copy_buffer_absolute_std (event, in_begins_absolute,
					  in_ends_absolute,
					  in_sizes,
					  in_adress,
					  in_elem_size,
					  out_begins,
					  out_ends,
					  out_sizes, out_adress,
					  out_elem_size, dest);
    }
}

#define sctk_elan_copy_message_func_one_get
static void
sctk_elan_copy_message_func (sctk_thread_ptp_message_t * restrict dest,
			     sctk_thread_ptp_message_t * restrict src)
{
  int i;

  void **adresses = NULL;
  sctk_pack_indexes_t **begins = NULL;
  sctk_pack_indexes_t **ends = NULL;
  sctk_pack_absolute_indexes_t **begins_absolute = NULL;
  sctk_pack_absolute_indexes_t **ends_absolute = NULL;
  size_t *sizes;
  size_t *elem_sizes;
  void **tmp_begins = NULL;
  void **tmp_ends = NULL;
  size_t max_count;
  size_t max_nb_items;

  char *tmp;
  char *current;
  size_t require = 0;

  ELAN_EVENT *event = NULL;
  static int done = 1;
  int real_nb_items;

  sctk_elan_ptp_init_put_get ();
/*   sctk_elan_lock(); */
  assume (dest->message.nb_items == src->message.nb_items);

  max_count = src->message.max_count;

#ifdef sctk_elan_copy_message_func_one_get
  max_nb_items = src->message.nb_items_max;
#else
  max_nb_items = src->message.nb_items + 1;
#endif

  require_malloc_ptp (max_nb_items * sizeof (sctk_pack_indexes_t *));
  require_malloc_ptp (max_nb_items * sizeof (sctk_pack_indexes_t *));
  require_malloc_ptp (max_nb_items * sizeof (sctk_pack_absolute_indexes_t *));
  require_malloc_ptp (max_nb_items * sizeof (sctk_pack_absolute_indexes_t *));
  require_malloc_ptp (max_nb_items * sizeof (size_t));
  require_malloc_ptp (max_nb_items * sizeof (size_t));
  require_malloc_ptp (max_nb_items * sizeof (void *));

  require_malloc_ptp (max_count * sizeof (void *));
  require_malloc_ptp (max_count * sizeof (void *));

  tmp = sctk_malloc (require);
  memset (tmp, 0, require);
  current = tmp;

  get_malloc_ptp (begins, max_nb_items * sizeof (sctk_pack_indexes_t *));
  get_malloc_ptp (ends, max_nb_items * sizeof (sctk_pack_indexes_t *));
  get_malloc_ptp (begins_absolute,
		  max_nb_items * sizeof (sctk_pack_absolute_indexes_t *));
  get_malloc_ptp (ends_absolute,
		  max_nb_items * sizeof (sctk_pack_absolute_indexes_t *));
  get_malloc_ptp (sizes, max_nb_items * sizeof (size_t));
  get_malloc_ptp (elem_sizes, max_nb_items * sizeof (size_t));
  get_malloc_ptp (adresses, max_nb_items * sizeof (void *));

  get_malloc_ptp (tmp_begins, max_count * sizeof (void *));
  get_malloc_ptp (tmp_ends, max_count * sizeof (void *));

#ifdef sctk_elan_copy_message_func_one_get
  {
    size_t message_get_size;

    message_get_size = SCTK_MESSAGE_UNIT_SIZE (src->message.nb_items_max);
    message_get_size -= sctk_aligned_sizeof (sctk_thread_ptp_message_t);

    sctk_nodebug ("get %lu", message_get_size);

    event = elan_doget ((ELAN_PGCTRL *) PGCTRL_PTP,
			src->message.begins, begins,
			message_get_size, src->process_src, ELAN_RAIL_ALL);
  }
#else
  event = elan_link (event, elan_doget ((ELAN_PGCTRL *) PGCTRL_PTP,
					src->message.begins, begins,
					max_nb_items *
					sizeof (sctk_pack_indexes_t *),
					src->process_src, ELAN_RAIL_ALL));
  event =
    elan_link (event,
	       elan_doget ((ELAN_PGCTRL *) PGCTRL_PTP, src->message.ends,
			   ends,
			   max_nb_items * sizeof (sctk_pack_indexes_t *),
			   src->process_src, ELAN_RAIL_ALL));
  event =
    elan_link (event,
	       elan_doget ((ELAN_PGCTRL *) PGCTRL_PTP,
			   src->message.begins_absolute, begins_absolute,
			   max_nb_items *
			   sizeof (sctk_pack_absolute_indexes_t *),
			   src->process_src, ELAN_RAIL_ALL));
  event =
    elan_link (event,
	       elan_doget ((ELAN_PGCTRL *) PGCTRL_PTP,
			   src->message.ends_absolute, ends_absolute,
			   max_nb_items *
			   sizeof (sctk_pack_absolute_indexes_t *),
			   src->process_src, ELAN_RAIL_ALL));
  event =
    elan_link (event,
	       elan_doget ((ELAN_PGCTRL *) PGCTRL_PTP, src->message.sizes,
			   sizes, max_nb_items * sizeof (size_t),
			   src->process_src, ELAN_RAIL_ALL));
  event =
    elan_link (event,
	       elan_doget ((ELAN_PGCTRL *) PGCTRL_PTP,
			   src->message.elem_sizes, elem_sizes,
			   max_nb_items * sizeof (size_t), src->process_src,
			   ELAN_RAIL_ALL));
  event =
    elan_link (event,
	       elan_doget ((ELAN_PGCTRL *) PGCTRL_PTP, src->message.adresses,
			   adresses, max_nb_items * sizeof (void *),
			   src->process_src, ELAN_RAIL_ALL));
#endif
  elan_wait (event, ELAN_WAIT_EVENT);
  event = NULL;

  real_nb_items = dest->message.nb_items;

  for (i = 0; i < real_nb_items; i++)
    {
      if (begins[i] != NULL)
	{
	  assume (sizes[i] <= max_count);
	  event = elan_link (event, elan_doget ((ELAN_PGCTRL *) PGCTRL_PTP,
						begins[i], tmp_begins,
						sizes[i] *
						sizeof (sctk_pack_indexes_t),
						src->process_src,
						ELAN_RAIL_ALL));
	  event =
	    elan_link (event,
		       elan_doget ((ELAN_PGCTRL *) PGCTRL_PTP, ends[i],
				   tmp_ends,
				   sizes[i] * sizeof (sctk_pack_indexes_t),
				   src->process_src, ELAN_RAIL_ALL));
	  begins[i] = (sctk_pack_indexes_t *) tmp_begins;
	  ends[i] = (sctk_pack_indexes_t *) tmp_ends;
	}
      else if (begins_absolute[i] != NULL)
	{
	  assume (sizes[i] <= max_count);
	  event = elan_link (event, elan_doget ((ELAN_PGCTRL *) PGCTRL_PTP,
						begins_absolute[i],
						tmp_begins,
						sizes[i] *
						sizeof
						(sctk_pack_absolute_indexes_t),
						src->process_src,
						ELAN_RAIL_ALL));
	  event =
	    elan_link (event,
		       elan_doget ((ELAN_PGCTRL *) PGCTRL_PTP,
				   ends_absolute[i], tmp_ends,
				   sizes[i] *
				   sizeof (sctk_pack_absolute_indexes_t),
				   src->process_src, ELAN_RAIL_ALL));
	  begins_absolute[i] = (sctk_pack_absolute_indexes_t *) tmp_begins;
	  ends_absolute[i] = (sctk_pack_absolute_indexes_t *) tmp_ends;
	}

      if (event)
	{
	  elan_wait (event, ELAN_WAIT_EVENT);
	  event = NULL;
	}

      sctk_elan_copy_buffer (&event,
			     begins[i],
			     ends[i],
			     begins_absolute[i],
			     ends_absolute[i],
			     sizes[i],
			     adresses[i],
			     elem_sizes[i],
			     dest->message.begins[i],
			     dest->message.ends[i],
			     dest->message.begins_absolute[i],
			     dest->message.ends_absolute[i],
			     dest->message.sizes[i],
			     dest->message.adresses[i],
			     dest->message.elem_sizes[i], src->process_src);
    }
  sctk_nodebug ("Free %p", src);

  if (event)
    {
      elan_wait (event, ELAN_WAIT_EVENT);
      event = NULL;
    }

  event =
    elan_link (event,
	       elan_doput ((ELAN_PGCTRL *) PGCTRL_PTP, &done,
			   src->addr_to_free,
			   (ELAN_ADDR) NULL,
			   sizeof (int), src->process_src, ELAN_RAIL_ALL));

  sctk_net_ptp_message_done (src);
  sctk_free (tmp);

/*   sctk_elan_unlock(); */
  elan_wait (event, ELAN_WAIT_EVENT);
}

static void
sctk_net_free_func_driver (sctk_thread_ptp_message_t * item)
{
  sctk_net_ptp_message_done (item);
}

/******************************************************************/
/* DSM                                                            */
/******************************************************************/
static ELAN_PGCTRL *PGCTRL_DSM;
static void
sctk_elan_get_pages_driver (void *addr, size_t size, int process)
{
  ELAN_EVENT *event;
  event = elan_doget (PGCTRL_DSM, addr, addr, size, process, ELAN_RAIL_ALL);
  elan_wait (event, ELAN_WAIT_EVENT);
}

/******************************************************************/
/* INIT                                                           */
/******************************************************************/
static void
sctk_elan_check (void *addr)
{
  if (addr)
    {
      int i;
      static void *addr_local;
      static void *addr_distant;

      addr_local = addr;
      addr_distant = addr;

      elan_gsync (elan_base->allGroup);
      for (i = 0; i < sctk_process_number; i++)
	{
	  elan_bcast (elan_base->allGroup, &addr_distant, sizeof (void *), i,
		      0);
	  assume (addr_local == addr_distant);
	}
      elan_gsync (elan_base->allGroup);
    }
}

static void *
sctk_elan_check_main (void *addr)
{
  int i;
  static void *addr_local;
  static void *addr_distant;
  int error = 0;

  addr_local = addr;
  addr_distant = addr;

  elan_gsync (elan_base->allGroup);
  for (i = 0; i < sctk_process_number; i++)
    {
      addr_distant = addr_local;
      elan_bcast (elan_base->allGroup, &addr_distant, sizeof (void *), i, 0);
      if (addr_local != addr_distant)
	{
	  if (sctk_process_rank == 0)
	    {
	      sctk_error ("ALIGN ERROR %d %p != %p", error, addr_local,
			  addr_distant);
	    }
	  error++;
	}
    }
  elan_gsync (elan_base->allGroup);
  assume (error == 0);
  return addr_local;
}
static char *sctk_elan_register_alloc_buff = NULL;
static size_t sctk_elan_register_alloc_buff_size = 0;
static size_t sctk_elan_register_alloc_buff_pos = 0;

static void *
sctk_elan_register_alloc (size_t size)
{
  char *tmp;
  assume (sctk_elan_register_alloc_buff_pos <=
	  sctk_elan_register_alloc_buff_size);
  tmp = sctk_elan_register_alloc_buff + sctk_elan_register_alloc_buff_pos;
  sctk_elan_register_alloc_buff_pos += sctk_aligned_size (size);
  return tmp;
}

static void *
sctk_elan_register_prealloc (size_t size)
{
  assume (sctk_elan_register_alloc_buff == NULL);

  size = sctk_aligned_size (size);
  sctk_elan_register_alloc_buff_size += size;
  return NULL;
}

static void
sctk_elan_register_prealloc_done ()
{
  size_t size;
  assume (sctk_elan_register_alloc_buff == NULL);

  size = sctk_elan_register_alloc_buff_size;
  sctk_elan_alloc_mem_used += size;
  sctk_elan_register_alloc_buff = sctk_iso_malloc (size);

  sctk_elan_register_alloc_buff =
    sctk_elan_check_main (sctk_elan_register_alloc_buff);

  memset (sctk_elan_register_alloc_buff, 0, size);
  sctk_elan_register_alloc_buff_size = size;
  sctk_elan_register_alloc_buff_pos = 0;

  sctk_nodebug ("Elan addresses %p", sctk_elan_register_alloc_buff);

}

static void *
sctk_elan_malloc (size_t size)
{
  if (sctk_elan_register_alloc_buff != NULL)
    {
      void *tmp;
      tmp = sctk_elan_register_alloc (size);
      memset (tmp, 0, size);
      sctk_elan_check (tmp);
      return tmp;
    }
  else
    {
      sctk_elan_register_prealloc (size);
      return NULL;
    }
}



static void
sctk_init_elan ()
{
  int i;

  /*Allocations */
  sctk_rpc_buffer = sctk_elan_malloc (sctk_process_number * sizeof (void *));
  for (i = 0; i < sctk_process_number; i++)
    {
      sctk_rpc_buffer[i] = sctk_elan_malloc (sctk_get_max_rpc_size_comm ());
    }
  sctk_send_rpc_used =
    sctk_elan_malloc (sctk_process_number * sizeof (sctk_rpc_t));
  sctk_recv_rpc_used =
    sctk_elan_malloc (sctk_process_number * sizeof (sctk_rpc_t));
  sctk_elan_ptp =
    sctk_elan_malloc (sctk_process_number * sizeof (sctk_elan_ptp_t));
  assume (sctk_elan_register_alloc_buff_pos ==
	  sctk_elan_register_alloc_buff_size);

  /*Initialisations */
  for (i = 0; i < sctk_process_number; i++)
    {
      sctk_elan_ptp_init (&(sctk_elan_ptp[i]), i);
    }
  elan_gsync (elan_base->allGroup);
  for (i = 0; i < SCTK_MAX_COMMUNICATOR_NUMBER; i++)
    {
      sctk_collective_init (&(sctk_bar[i]), sctk_process_number);
    }

  elan_gsync (elan_base->allGroup);

  sctk_nodebug ("sctk_send_rpc_used %p-%p", sctk_send_rpc_used,
		(char *) sctk_send_rpc_used +
		sctk_process_number * sizeof (sctk_rpc_t));
  sctk_nodebug ("sctk_recv_rpc_used %p-%p", sctk_recv_rpc_used,
		(char *) sctk_recv_rpc_used +
		sctk_process_number * sizeof (sctk_rpc_t));

  sctk_nodebug ("ELAN init done %d", sctk_process_rank);
  if (sctk_process_rank == 0)
    {
      init_elan_display ();
    }
  elan_gsync (elan_base->allGroup);

  PGCTRL_RPC = elan_putgetInit (elan_base->state, NULL,
				elan_base->putget_smallputsize,
				elan_base->putget_stripeputsize,
				elan_base->putget_stripegetsize,
				elan_base->putget_throttle,
				elan_base->putget_flags);

  PGCTRL_COLLECTIVE = elan_putgetInit (elan_base->state, NULL,
				       elan_base->putget_smallputsize,
				       elan_base->putget_stripeputsize,
				       elan_base->putget_stripegetsize,
				       elan_base->putget_throttle,
				       elan_base->putget_flags);

  PGCTRL_DSM = elan_putgetInit (elan_base->state, NULL,
				elan_base->putget_smallputsize,
				elan_base->putget_stripeputsize,
				elan_base->putget_stripegetsize,
				elan_base->putget_throttle,
				elan_base->putget_flags);

  elan_gsync (elan_base->allGroup);
  sctk_elan_ptp_init_put_get ();

  elan_gsync (elan_base->allGroup);
  sctk_launch_poll_thread ();
  elan_gsync (elan_base->allGroup);
}

void
sctk_net_init_driver_elan4 (int *argc, char ***argv)
{
  assume (argc != NULL);
  assume (argv != NULL);
  sctk_rpc_init_func (sctk_net_rpc_driver, sctk_net_rpc_retrive_driver,
		      sctk_net_rpc_send_driver);
  sctk_send_init_func (sctk_net_send_ptp_message_driver,
		       sctk_elan_copy_message_func,
		       sctk_net_free_func_driver);
  sctk_collective_init_func (sctk_net_collective_op_driver);
  sctk_net_get_pages_init (sctk_elan_get_pages_driver);

  sctk_init_elan ();

  sctk_net_adm_poll = sctk_net_adm_poll_elan;
  sctk_net_ptp_poll = sctk_net_ptp_poll_elan;
  sctk_nodebug ("ELAN init done");
  elan_gsync (elan_base->allGroup);
}

void
sctk_net_preinit_driver_elan4 ()
{
  int i;
  sctk_nodebug ("%d\n", getpid ());

  elan_base = elan_baseInit (0);

  sctk_elan_alloc_mem_used += sizeof (sctk_bar);
  sctk_process_rank = elan_base->state->vp;
  sctk_process_number = elan_base->state->nvp;

  elan_gsync (elan_base->allGroup);

  /*PRE Allocations */
  sctk_rpc_buffer = sctk_elan_malloc (sctk_process_number * sizeof (void *));
  for (i = 0; i < sctk_process_number; i++)
    {
      sctk_elan_malloc (sctk_get_max_rpc_size_comm ());
    }
  sctk_send_rpc_used =
    sctk_elan_malloc (sctk_process_number * sizeof (sctk_rpc_t));
  sctk_recv_rpc_used =
    sctk_elan_malloc (sctk_process_number * sizeof (sctk_rpc_t));
  sctk_elan_ptp =
    sctk_elan_malloc (sctk_process_number * sizeof (sctk_elan_ptp_t));

  sctk_elan_register_prealloc_done ();
  elan_gsync (elan_base->allGroup);
#ifdef ELAN_LOG
  do
    {
      static char name[4096];
      sprintf (name, "elan_log_file_%05d", sctk_process_rank);
      elan_log_file = fopen (name, "w");
      sctk_log (elan_log_file, "Start_logging");
    }
  while (0);
#endif
}
#else
void
sctk_net_init_driver_elan4 (int *argc, char ***argv)
{
  assume (argc != NULL);
  assume (argv != NULL);
  not_available ();
}

void
sctk_net_preinit_driver_elan4 ()
{
  not_available ();
}
#endif
