/* ############################# MPC License ############################## */
/* # Wed Nov 19 15:19:19 CET 2008                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # IDDN.FR.001.230040.000.S.P.2007.000.10000                            # */
/* # This file is part of the MPC Runtime.                                # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - PERACHE Marc marc.perache@cea.fr                                 # */
/* #                                                                      # */
/* ######################################################################## */
#include "sctk_low_level_comm.h"
#include "sctk_debug.h"
#include "sctk_thread.h"
#include "sctk_kernel_thread.h"
#include "sctk_rpc.h"
#include "sctk_elan_gen.h"
#include <math.h>
#include <sched.h>
#include <sctk_asm.h>

#ifdef MPC_USE_ELAN
#include "sctk_net_tools.h"
#include "sctk_iso_alloc.h"

#define SCTK_MAX_RAIL_NUMBER 3
/* #define ELAN_LOG */

#include <elan/elan.h>

#if !defined(_ELAN_DEVENT_H) || !defined(ELAN_EVENT_ALIGN)

typedef struct
{
  char data[32];
} ELAN_EVENT_ELAN;

#define ELAN_EVENT_ALIGN 64
extern ELAN_EVENT *elan_initEvent (ELAN_STATE * state, ELAN_RAIL * r,
				   ELAN_ADDR elan, int count);
extern ELAN_ADDR elan_destEvent (ELAN_EVENT * e);
extern void elan_issueWaitevent (ELAN_EVENT * e, int count);
extern void _elan_deventDump (const char *lable, ELAN_EVENT * e);
#endif


ELAN_BASE *elan_base;
int num_devices;
/**************************************************************************/
/* RDMA DRIVER                                                            */
/**************************************************************************/

/*HAVE distinguish remote and lock acks*/
typedef struct
{
  volatile int done;
  ELAN_EVENT *event;
} sctk_net_rdma_ack_t;

static int
sctk_net_rdma_write (void *local_addr, void *dist_addr, size_t size,
		     int process, sctk_net_rdma_ack_t * local_ack,
		     sctk_net_rdma_ack_t * dist_ack)
{
  if (local_ack == NULL)
    {
      elan_wait (elan_put
		 (elan_base->state, local_addr, dist_addr, size, process),
		 ELAN_WAIT_EVENT);
      if (dist_ack)
	{
	  dist_ack->done = 1;
	}
    }
  else
    {
      local_ack->event =
	elan_put (elan_base->state, local_addr, dist_addr, size, process);
      local_ack->done = 0;

      if (dist_ack)
	{
	  dist_ack->event = NULL;
	  dist_ack->done = 1;
	}
    }

  return 0;
}
static int
sctk_net_rdma_read (void *local_addr, void *dist_addr, size_t size,
		    int process, sctk_net_rdma_ack_t * ack)
{
  not_implemented ();
  return 0;
}

typedef struct
{
  sctk_net_rdma_ack_t *ack;
  int done;
} sctk_net_rdma_wait_t;

static void
sctk_net_rdma_wait_poll_func (sctk_net_rdma_wait_t * arg)
{
  int done;

  done = arg->ack->done;
  if (!done)
    {
      if (arg->ack->event)
	{
	  done = elan_poll (arg->ack->event, 1);
	}
      else
	{
	  done = 1;
	}
      if (done)
	{
	  arg->ack->done = 0;
	  arg->ack->event = NULL;
	  done = 1;
	}
    }
  arg->done = done;
}

static void
sctk_net_rdma_wait (sctk_net_rdma_ack_t * ack)
{
  sctk_net_rdma_wait_t wait_arg;
  int i;
  wait_arg.ack = ack;
  wait_arg.done = 0;
  sctk_thread_wait_for_value_and_poll (&(wait_arg.done), 1,
				       (void (*)(void *))
				       sctk_net_rdma_wait_poll_func,
				       &wait_arg);
}

static int
sctk_net_rdma_test (sctk_net_rdma_ack_t * ack)
{
  sctk_net_rdma_wait_t wait_arg;
  wait_arg.ack = ack;
  wait_arg.done = 0;

  sctk_net_rdma_wait_poll_func (&wait_arg);
  return wait_arg.done;
}

/******************************************************************/
/* RPC                                                            */
/******************************************************************/
#include "sctk_rpc_rdma.h"

/******************************************************************/
/* COLLECTIVES                                                    */
/******************************************************************/
#define collective_op_steps 32
#define collective_buffer_size (8*1024)
#include "sctk_collectives_rdma.h"

/******************************************************************/
/* PT 2 PT                                                        */
/******************************************************************/
#define SCTK_NET_MESSAGE_NUMBER 128
#define SCTK_NET_MESSAGE_BUFFER (16*1024*SCTK_NET_MESSAGE_NUMBER)
#define SCTK_NET_MESSAGE_SPLIT_THRESHOLD (32*1024)
#include "sctk_pt2pt_rdma.h"

/******************************************************************/
/* DSM                                                            */
/******************************************************************/
#include "sctk_dsm_rdma.h"

/******************************************************************/
/* INIT/END                                                       */
/******************************************************************/
#include "sctk_init_rdma.h"

void
sctk_net_init_driver_elan_gen (int *argc, char ***argv)
{
  assume (argc != NULL);
  assume (argv != NULL);

  sctk_net_init_driver_init_modules ("ELAN GEN");

  elan_gsync (elan_base->allGroup);
}

void
sctk_net_preinit_driver_elan_gen ()
{
  elan_base = elan_baseInit (0);

  sctk_process_rank = elan_base->state->vp;
  sctk_process_number = elan_base->state->nvp;

  num_devices = elan_nRails (elan_base->state);
  assume (num_devices <= SCTK_MAX_RAIL_NUMBER);

  elan_gsync (elan_base->allGroup);

  sctk_net_preinit_driver_init_modules ();

  elan_gsync (elan_base->allGroup);
}
#else
void
sctk_net_init_driver_elan_gen (int *argc, char ***argv)
{
  assume (argc != NULL);
  assume (argv != NULL);
  not_available ();
}

void
sctk_net_preinit_driver_elan_gen ()
{
  not_available ();
}
#endif
